/* Copyright (C) 2017-2021 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Maksym Zyzak [committer], Valentina Akishina */

#include "CaTrackFit.h"

#include "CaHit.h"
#include "CaMeasurementU.h"
#include "CaMeasurementXy.h"
#include "CaStation.h"

namespace cbm::algo::ca
{

  typedef const fvec cnst;

  void TrackFit::Filter1d(const ca::MeasurementU<fvec>& m)
  {
    fvec zeta, HCH;
    fvec F0, F1, F2, F3, F4, F5, F6;
    fvec K1, K2, K3, K4, K5, K6;

    zeta = m.CosPhi() * fTr.X() + m.SinPhi() * fTr.Y() - m.U();

    // F = CH'
    F0 = m.CosPhi() * fTr.C00() + m.SinPhi() * fTr.C10();
    F1 = m.CosPhi() * fTr.C10() + m.SinPhi() * fTr.C11();

    HCH = (F0 * m.CosPhi() + F1 * m.SinPhi());

    F2 = m.CosPhi() * fTr.C20() + m.SinPhi() * fTr.C21();
    F3 = m.CosPhi() * fTr.C30() + m.SinPhi() * fTr.C31();
    F4 = m.CosPhi() * fTr.C40() + m.SinPhi() * fTr.C41();
    F5 = m.CosPhi() * fTr.C50() + m.SinPhi() * fTr.C51();
    F6 = m.CosPhi() * fTr.C60() + m.SinPhi() * fTr.C61();

    const fmask maskDoFilter = (HCH < m.Du2() * 16.f);
    //cnst maskDoFilter = _f32vec4_true;

    // correction to HCH is needed for the case when sigma2 is so small
    // with respect to HCH that it disappears due to the roundoff error
    //
    fvec wi     = fMaskF / (m.Du2() + fvec(1.0000001) * HCH);
    fvec zetawi = fMaskF * zeta / (iif(maskDoFilter, m.Du2(), fvec::Zero()) + HCH);

    //wi = iif(wi > fvec::Zero(), wi, fvec::Zero());
    wi = iif(m.Du2() > fvec::Zero(), wi, fvec::Zero());

    // fTr.ChiSq() += iif( maskDoFilter, zeta * zetawi, fvec::Zero() );
    fTr.ChiSq() += m.Ndf() * zeta * zeta * wi;
    fTr.Ndf() += m.Ndf() * fMaskF;

    K1 = F1 * wi;
    K2 = F2 * wi;
    K3 = F3 * wi;
    K4 = F4 * wi;
    K5 = F5 * wi;
    K6 = F6 * wi;

    fTr.X() -= F0 * zetawi;
    fTr.Y() -= F1 * zetawi;
    fTr.Tx() -= F2 * zetawi;
    fTr.Ty() -= F3 * zetawi;
    fTr.Qp() -= F4 * zetawi;
    fTr.Time() -= F5 * zetawi;
    fTr.Vi() -= F6 * zetawi;

    fTr.C00() -= F0 * F0 * wi;

    fTr.C10() -= K1 * F0;
    fTr.C11() -= K1 * F1;

    fTr.C20() -= K2 * F0;
    fTr.C21() -= K2 * F1;
    fTr.C22() -= K2 * F2;

    fTr.C30() -= K3 * F0;
    fTr.C31() -= K3 * F1;
    fTr.C32() -= K3 * F2;
    fTr.C33() -= K3 * F3;

    fTr.C40() -= K4 * F0;
    fTr.C41() -= K4 * F1;
    fTr.C42() -= K4 * F2;
    fTr.C43() -= K4 * F3;
    fTr.C44() -= K4 * F4;

    fTr.C50() -= K5 * F0;
    fTr.C51() -= K5 * F1;
    fTr.C52() -= K5 * F2;
    fTr.C53() -= K5 * F3;
    fTr.C54() -= K5 * F4;
    fTr.C55() -= K5 * F5;

    fTr.C60() -= K6 * F0;
    fTr.C61() -= K6 * F1;
    fTr.C62() -= K6 * F2;
    fTr.C63() -= K6 * F3;
    fTr.C64() -= K6 * F4;
    fTr.C65() -= K6 * F5;
    fTr.C66() -= K6 * F6;
  }

  void TrackFit::FilterTime(fvec t, fvec dt2, fvec timeInfo)
  {
    // filter track with a time measurement

    // F = CH'
    fvec F0 = fTr.C50();
    fvec F1 = fTr.C51();
    fvec F2 = fTr.C52();
    fvec F3 = fTr.C53();
    fvec F4 = fTr.C54();
    fvec F5 = fTr.C55();
    fvec F6 = fTr.C65();

    fvec HCH = fTr.C55();

    fvec w = fMaskF;

    w.setZero(timeInfo <= fvec::Zero());

    // when dt0 is much smaller than current time error,
    // set track time exactly to the measurement value without filtering
    // it helps to keep the initial time errors reasonably small
    // the calculations in the covariance matrix are not affected

    const fmask maskDoFilter = (HCH < dt2 * 16.f);

    fvec wi     = w / (dt2 + fvec(1.0000001) * HCH);
    fvec zeta   = fTr.Time() - t;
    fvec zetawi = w * zeta / (iif(maskDoFilter, dt2, fvec::Zero()) + HCH);

    fTr.ChiSqTime()(maskDoFilter) += zeta * zeta * wi;
    fTr.NdfTime() += w;

    fvec K1 = F1 * wi;
    fvec K2 = F2 * wi;
    fvec K3 = F3 * wi;
    fvec K4 = F4 * wi;
    fvec K5 = F5 * wi;
    fvec K6 = F6 * wi;

    fTr.X() -= F0 * zetawi;
    fTr.Y() -= F1 * zetawi;
    fTr.Tx() -= F2 * zetawi;
    fTr.Ty() -= F3 * zetawi;
    fTr.Qp() -= F4 * zetawi;
    fTr.Time() -= F5 * zetawi;
    fTr.Vi() -= F6 * zetawi;

    fTr.C00() -= F0 * F0 * wi;

    fTr.C10() -= K1 * F0;
    fTr.C11() -= K1 * F1;

    fTr.C20() -= K2 * F0;
    fTr.C21() -= K2 * F1;
    fTr.C22() -= K2 * F2;

    fTr.C30() -= K3 * F0;
    fTr.C31() -= K3 * F1;
    fTr.C32() -= K3 * F2;
    fTr.C33() -= K3 * F3;

    fTr.C40() -= K4 * F0;
    fTr.C41() -= K4 * F1;
    fTr.C42() -= K4 * F2;
    fTr.C43() -= K4 * F3;
    fTr.C44() -= K4 * F4;

    fTr.C50() -= K5 * F0;
    fTr.C51() -= K5 * F1;
    fTr.C52() -= K5 * F2;
    fTr.C53() -= K5 * F3;
    fTr.C54() -= K5 * F4;
    fTr.C55() -= K5 * F5;

    fTr.C60() -= K6 * F0;
    fTr.C61() -= K6 * F1;
    fTr.C62() -= K6 * F2;
    fTr.C63() -= K6 * F3;
    fTr.C64() -= K6 * F4;
    fTr.C65() -= K6 * F5;
    fTr.C66() -= K6 * F6;
  }


  void TrackFit::FilterXY(const ca::MeasurementXy<fvec>& mxy)
  {
    {
      ca::MeasurementU<fvec> mx;
      mx.SetCosPhi(fvec::One());
      mx.SetSinPhi(fvec::Zero());
      mx.SetU(mxy.X());
      mx.SetDu2(mxy.Dx2());
      mx.SetNdf(mxy.NdfX());

      ca::MeasurementU<fvec> mu;
      mu.SetCosPhi(-mxy.Dxy() / mxy.Dx2());
      mu.SetSinPhi(fvec::One());
      mu.SetU(mu.CosPhi() * mxy.X() + mxy.Y());
      mu.SetDu2(mxy.Dy2() - mxy.Dxy() * mxy.Dxy() / mxy.Dx2());
      mu.SetNdf(mxy.NdfY());

      Filter1d(mx);
      Filter1d(mu);
      return;
    }

    //----------------------------------------------------------------------------------------------
    // the other way: filter 2 dimensions at once

    cnst TWO(2.);

    fvec zeta0, zeta1, S00, S10, S11, si;
    fvec F00, F10, F20, F30, F40, F50, F60;
    fvec F01, F11, F21, F31, F41, F51, F61;
    fvec K00, K10, K20, K30, K40, K50, K60;
    fvec K01, K11, K21, K31, K41, K51, K61;

    zeta0 = fTr.X() - mxy.X();
    zeta1 = fTr.Y() - mxy.Y();

    // F = CH'
    F00 = fTr.C00();
    F10 = fTr.C10();
    F20 = fTr.C20();
    F30 = fTr.C30();
    F40 = fTr.C40();
    F50 = fTr.C50();
    F60 = fTr.C60();

    F01 = fTr.C10();
    F11 = fTr.C11();
    F21 = fTr.C21();
    F31 = fTr.C31();
    F41 = fTr.C41();
    F51 = fTr.C51();
    F61 = fTr.C61();

    S00 = F00 + mxy.Dx2();
    S10 = F10 + mxy.Dxy();
    S11 = F11 + mxy.Dy2();

    si          = 1.f / (S00 * S11 - S10 * S10);
    fvec S00tmp = S00;
    S00         = si * S11;
    S10         = -si * S10;
    S11         = si * S00tmp;

    fTr.ChiSq() += zeta0 * zeta0 * S00 + 2.f * zeta0 * zeta1 * S10 + zeta1 * zeta1 * S11;
    fTr.Ndf() += TWO;

    K00 = F00 * S00 + F01 * S10;
    K01 = F00 * S10 + F01 * S11;

    K10 = F10 * S00 + F11 * S10;
    K11 = F10 * S10 + F11 * S11;

    K20 = F20 * S00 + F21 * S10;
    K21 = F20 * S10 + F21 * S11;

    K30 = F30 * S00 + F31 * S10;
    K31 = F30 * S10 + F31 * S11;

    K40 = F40 * S00 + F41 * S10;
    K41 = F40 * S10 + F41 * S11;

    K50 = F50 * S00 + F51 * S10;
    K51 = F50 * S10 + F51 * S11;

    K60 = F60 * S00 + F61 * S10;
    K61 = F60 * S10 + F61 * S11;

    fTr.X() -= K00 * zeta0 + K01 * zeta1;
    fTr.Y() -= K10 * zeta0 + K11 * zeta1;
    fTr.Tx() -= K20 * zeta0 + K21 * zeta1;
    fTr.Ty() -= K30 * zeta0 + K31 * zeta1;
    fTr.Qp() -= K40 * zeta0 + K41 * zeta1;
    fTr.Time() -= K50 * zeta0 + K51 * zeta1;
    fTr.Vi() -= K60 * zeta0 + K61 * zeta1;

    fTr.C00() -= K00 * F00 + K01 * F01;

    fTr.C10() -= K10 * F00 + K11 * F01;
    fTr.C11() -= K10 * F10 + K11 * F11;

    fTr.C20() -= K20 * F00 + K21 * F01;
    fTr.C21() -= K20 * F10 + K21 * F11;
    fTr.C22() -= K20 * F20 + K21 * F21;

    fTr.C30() -= K30 * F00 + K31 * F01;
    fTr.C31() -= K30 * F10 + K31 * F11;
    fTr.C32() -= K30 * F20 + K31 * F21;
    fTr.C33() -= K30 * F30 + K31 * F31;

    fTr.C40() -= K40 * F00 + K41 * F01;
    fTr.C41() -= K40 * F10 + K41 * F11;
    fTr.C42() -= K40 * F20 + K41 * F21;
    fTr.C43() -= K40 * F30 + K41 * F31;
    fTr.C44() -= K40 * F40 + K41 * F41;

    fTr.C50() -= K50 * F00 + K51 * F01;
    fTr.C51() -= K50 * F10 + K51 * F11;
    fTr.C52() -= K50 * F20 + K51 * F21;
    fTr.C53() -= K50 * F30 + K51 * F31;
    fTr.C54() -= K50 * F40 + K51 * F41;
    fTr.C55() -= K50 * F50 + K51 * F51;

    fTr.C60() -= K60 * F00 + K61 * F01;
    fTr.C61() -= K60 * F10 + K61 * F11;
    fTr.C62() -= K60 * F20 + K61 * F21;
    fTr.C63() -= K60 * F30 + K61 * F31;
    fTr.C64() -= K60 * F40 + K61 * F41;
    fTr.C65() -= K60 * F50 + K61 * F51;
    fTr.C66() -= K60 * F60 + K61 * F61;
  }

  void TrackFit::FilterHit(const ca::Station<fvec>& sta, const ca::Hit& hit)
  {
    ca::MeasurementXy<fvec> m;
    m.SetDx2(hit.dX2());
    m.SetDy2(hit.dY2());
    m.SetDxy(hit.dXY());
    m.SetX(hit.X());
    m.SetY(hit.Y());
    m.SetNdfX(fvec::One());
    m.SetNdfY(fvec::One());
    FilterXY(m);
    FilterTime(hit.T(), hit.dT2(), sta.timeInfo);
  }

  void TrackFit::FilterExtrapolatedXY(const ca::MeasurementXy<fvec>& m, fvec extrX, fvec extrY,
                                      const std::array<fvec, ca::TrackParamV::kNtrackParam>& Jx,
                                      const std::array<fvec, ca::TrackParamV::kNtrackParam>& Jy)
  {
    // add a 2-D measurenent (x,y) at some z, that differs from fTr.GetZ()
    // extrX, extrY are extrapolated track parameters at z, Jx, Jy are derivatives of the extrapolation

    // ! it is assumed that in the track covariance matrix all non-diagonal covariances are 0
    // ! except of C10

    TrackParamV& T = fTr;

    //zeta0 = T.x + Jx[2]*T.Tx() + Jx[3]*T.Ty() + Jx[4]*T.qp - x;
    //zeta1 = T.y + Jy[2]*T.Tx() + Jy[3]*T.Ty() + Jy[4]*T.qp - y;

    fvec zeta0 = extrX - m.X();
    fvec zeta1 = extrY - m.Y();

    // H = 1 0 Jx[2] Jx[3] Jx[4] 0
    //     0 1 Jy[2] Jy[3] Jy[4] 0

    // F = CH'
    fvec F00 = T.C00();
    fvec F01 = T.C10();
    fvec F10 = T.C10();
    fvec F11 = T.C11();
    fvec F20 = Jx[2] * T.C22();
    fvec F21 = Jy[2] * T.C22();
    fvec F30 = Jx[3] * T.C33();
    fvec F31 = Jy[3] * T.C33();
    fvec F40 = Jx[4] * T.C44();
    fvec F41 = Jy[4] * T.C44();

    // Jx[5,6] and Jy[5,6] are 0.

    fvec S00 = m.Dx2() + F00 + Jx[2] * F20 + Jx[3] * F30 + Jx[4] * F40;
    fvec S10 = m.Dxy() + F10 + Jy[2] * F20 + Jy[3] * F30 + Jy[4] * F40;
    fvec S11 = m.Dy2() + F11 + Jy[2] * F21 + Jy[3] * F31 + Jy[4] * F41;

    fvec si = fvec(1.) / (S00 * S11 - S10 * S10);

    fvec S00tmp = S00;
    S00         = si * S11;
    S10         = -si * S10;
    S11         = si * S00tmp;

    T.ChiSq() += zeta0 * zeta0 * S00 + fvec(2.) * zeta0 * zeta1 * S10 + zeta1 * zeta1 * S11;
    T.Ndf() += m.NdfX() + m.NdfY();

    fvec K00 = F00 * S00 + F01 * S10;
    fvec K01 = F00 * S10 + F01 * S11;
    fvec K10 = F10 * S00 + F11 * S10;
    fvec K11 = F10 * S10 + F11 * S11;
    fvec K20 = F20 * S00 + F21 * S10;
    fvec K21 = F20 * S10 + F21 * S11;
    fvec K30 = F30 * S00 + F31 * S10;
    fvec K31 = F30 * S10 + F31 * S11;
    fvec K40 = F40 * S00 + F41 * S10;
    fvec K41 = F40 * S10 + F41 * S11;

    T.X() -= K00 * zeta0 + K01 * zeta1;
    T.Y() -= K10 * zeta0 + K11 * zeta1;
    T.Tx() -= K20 * zeta0 + K21 * zeta1;
    T.Ty() -= K30 * zeta0 + K31 * zeta1;
    T.Qp() -= K40 * zeta0 + K41 * zeta1;

    T.C00() -= (K00 * F00 + K01 * F01);
    T.C10() -= (K10 * F00 + K11 * F01);
    T.C11() -= (K10 * F10 + K11 * F11);
    T.C20() = -(K20 * F00 + K21 * F01);
    T.C21() = -(K20 * F10 + K21 * F11);
    T.C22() -= (K20 * F20 + K21 * F21);
    T.C30() = -(K30 * F00 + K31 * F01);
    T.C31() = -(K30 * F10 + K31 * F11);
    T.C32() = -(K30 * F20 + K31 * F21);
    T.C33() -= (K30 * F30 + K31 * F31);
    T.C40() = -(K40 * F00 + K41 * F01);
    T.C41() = -(K40 * F10 + K41 * F11);
    T.C42() = -(K40 * F20 + K41 * F21);
    T.C43() = -(K40 * F30 + K41 * F31);
    T.C44() -= (K40 * F40 + K41 * F41);
  }


  void TrackFit::MeasureVelocityWithQp()
  {
    // measure velocity using measured qp
    // assuming particle mass == fMass;

    cnst kClightNsInv = fvec(constants::phys::SpeedOfLightInv);

    fvec zeta, HCH;
    fvec F0, F1, F2, F3, F4, F5, F6;
    fvec K1, K2, K3, K4, K5, K6;

    //FilterVi(sqrt(fvec(1.) + fMass2 * fQp0 * fQp0) * kClightNsInv);
    //return;

    fvec vi0 = sqrt(fvec(1.) + fMass2 * fQp0 * fQp0) * kClightNsInv;

    fvec h = fMass2 * fQp0 / sqrt(fvec(1.) + fMass2 * fQp0 * fQp0) * kClightNsInv;

    zeta = vi0 + h * (fTr.Qp() - fQp0) - fTr.Vi();

    fTr.Vi() = vi0;

    // H = (0,0,0,0, h,0, -1)

    // F = CH'

    F0 = h * fTr.C40() - fTr.C60();
    F1 = h * fTr.C41() - fTr.C61();
    F2 = h * fTr.C42() - fTr.C62();
    F3 = h * fTr.C43() - fTr.C63();
    F4 = h * fTr.C44() - fTr.C64();
    F5 = h * fTr.C54() - fTr.C65();
    F6 = h * fTr.C64() - fTr.C66();

    HCH = F4 * h - F6;

    fvec wi     = fMaskF / HCH;
    fvec zetawi = fMaskF * zeta / HCH;
    fTr.ChiSqTime() += zeta * zeta * wi;
    fTr.NdfTime() += fMaskF;

    K1 = F1 * wi;
    K2 = F2 * wi;
    K3 = F3 * wi;
    K4 = F4 * wi;
    K5 = F5 * wi;
    K6 = F6 * wi;

    fTr.X() -= F0 * zetawi;
    fTr.Y() -= F1 * zetawi;
    fTr.Tx() -= F2 * zetawi;
    fTr.Ty() -= F3 * zetawi;
    fTr.Qp() -= F4 * zetawi;
    fTr.Time() -= F5 * zetawi;
    fTr.Vi() -= F6 * zetawi;

    fTr.C00() -= F0 * F0 * wi;

    fTr.C10() -= K1 * F0;
    fTr.C11() -= K1 * F1;

    fTr.C20() -= K2 * F0;
    fTr.C21() -= K2 * F1;
    fTr.C22() -= K2 * F2;

    fTr.C30() -= K3 * F0;
    fTr.C31() -= K3 * F1;
    fTr.C32() -= K3 * F2;
    fTr.C33() -= K3 * F3;

    fTr.C40() -= K4 * F0;
    fTr.C41() -= K4 * F1;
    fTr.C42() -= K4 * F2;
    fTr.C43() -= K4 * F3;
    fTr.C44() -= K4 * F4;

    fTr.C50() -= K5 * F0;
    fTr.C51() -= K5 * F1;
    fTr.C52() -= K5 * F2;
    fTr.C53() -= K5 * F3;
    fTr.C54() -= K5 * F4;
    fTr.C55() -= K5 * F5;

    fTr.C60() -= K6 * F0;
    fTr.C61() -= K6 * F1;
    fTr.C62() -= K6 * F2;
    fTr.C63() -= K6 * F3;
    fTr.C64() -= K6 * F4;
    fTr.C65() -= K6 * F5;
    fTr.C66() -= K6 * F6;

    //  fTr.Vi()( fTr.Vi() < fvec(TrackParamV::kClightNsInv) ) = fvec(TrackParamV::kClightNsInv);
  }

  void TrackFit::FilterVi(fvec vi)
  {
    // set inverse velocity to vi

    fvec zeta, HCH;
    fvec F0, F1, F2, F3, F4, F5, F6;
    fvec K1, K2, K3, K4, K5, K6;

    zeta = fTr.Vi() - vi;

    // H = (0,0,0,0, 0, 0, 1)

    // F = CH'

    F0 = fTr.C60();
    F1 = fTr.C61();
    F2 = fTr.C62();
    F3 = fTr.C63();
    F4 = fTr.C64();
    F5 = fTr.C65();
    F6 = fTr.C66();

    HCH = F6;

    fvec wi     = fMaskF / HCH;
    fvec zetawi = fMaskF * zeta / HCH;
    fTr.ChiSqTime() += zeta * zeta * wi;
    fTr.NdfTime() += fMaskF;

    K1 = F1 * wi;
    K2 = F2 * wi;
    K3 = F3 * wi;
    K4 = F4 * wi;
    K5 = F5 * wi;
    K6 = F6 * wi;

    fTr.X() -= F0 * zetawi;
    fTr.Y() -= F1 * zetawi;
    fTr.Tx() -= F2 * zetawi;
    fTr.Ty() -= F3 * zetawi;
    fTr.Qp() -= F4 * zetawi;
    fTr.Time() -= F5 * zetawi;
    // fTr.Vi() -= F6 * zetawi;
    fTr.Vi() = vi;

    fTr.C00() -= F0 * F0 * wi;

    fTr.C10() -= K1 * F0;
    fTr.C11() -= K1 * F1;

    fTr.C20() -= K2 * F0;
    fTr.C21() -= K2 * F1;
    fTr.C22() -= K2 * F2;

    fTr.C30() -= K3 * F0;
    fTr.C31() -= K3 * F1;
    fTr.C32() -= K3 * F2;
    fTr.C33() -= K3 * F3;

    fTr.C40() -= K4 * F0;
    fTr.C41() -= K4 * F1;
    fTr.C42() -= K4 * F2;
    fTr.C43() -= K4 * F3;
    fTr.C44() -= K4 * F4;

    fTr.C50() -= K5 * F0;
    fTr.C51() -= K5 * F1;
    fTr.C52() -= K5 * F2;
    fTr.C53() -= K5 * F3;
    fTr.C54() -= K5 * F4;
    fTr.C55() -= K5 * F5;

    //fTr.C60() -= K6 * F0;
    //fTr.C61() -= K6 * F1;
    //fTr.C62() -= K6 * F2;
    //fTr.C63() -= K6 * F3;
    //fTr.C64() -= K6 * F4;
    //fTr.C65() -= K6 * F5;
    //fTr.C66() -= K6 * F6;
    fTr.C60() = fvec(0.);
    fTr.C61() = fvec(0.);
    fTr.C62() = fvec(0.);
    fTr.C63() = fvec(0.);
    fTr.C64() = fvec(0.);
    fTr.C65() = fvec(0.);
    fTr.C66() = fvec(1.e-8);  // just for a case..
  }

  void TrackFit::Extrapolate  // extrapolates track parameters and returns jacobian for extrapolation of CovMatrix
    (fvec z_out,              // extrapolate to this z position
     const ca::FieldRegion<fvec>& F)
  {
    // use Q/p linearisation at fQp0

    if (F.IsZeroField()) {
      ExtrapolateLineNoField(z_out);
    }
    else {
      fvec sgn = iif(fTr.GetZ() < z_out, fvec(1.), fvec(-1.));
      while (!(fMaskF * abs(z_out - fTr.GetZ()) <= fvec(1.e-6)).isFull()) {
        fvec zNew                              = fTr.GetZ() + sgn * fMaxExtraplationStep;  // max. 50 cm step
        zNew(sgn * (z_out - zNew) <= fvec(0.)) = z_out;
        ExtrapolateStep(zNew, F);
      }
    }
  }

  void TrackFit::ExtrapolateStep  // extrapolates track parameters and returns jacobian for extrapolation of CovMatrix
    (fvec zOut,                   // extrapolate to this z position
     const ca::FieldRegion<fvec>& Field)
  {
    // use Q/p linearisation at fQp0
    // implementation of the Runge-Kutta method without optimization
    //

    //
    // Forth-order Runge-Kutta method for solution of the equation
    // of motion of a particle with parameter qp = Q /P
    //              in the magnetic field B()
    //
    //   ( x )            tx
    //   ( y )            ty
    //   ( tx)        c_light * qp * L * (     tx*ty * Bx - (1+tx*tx) * By + ty * Bz  )
    // d ( ty) / dz = c_light * qp * L * ( (1+ty*ty) * Bx     - tx*ty * By - tx * Bz  )  ,
    //   ( qp)             0.
    //   ( t )         L * vi
    //   ( vi)             0.
    //
    //   where  L = sqrt ( 1 + tx*tx + ty*ty ) .
    //   c_light = 0.000299792458 [(GeV/c)/kG/cm]
    //   c_light_ns =  29.9792458 [cm/ns]
    //
    //  In the following for RK step :
    //   r[7] = {x, y, tx, ty, qp, t, vi}
    //   dr(z)/dz = f(z,r)
    //
    //
    //========================================================================
    //
    //  NIM A395 (1997) 169-184; NIM A426 (1999) 268-282.
    //
    //  the routine is based on LHC(b) utility code
    //
    //========================================================================


    cnst c_light = 0.000299792458;

    //----------------------------------------------------------------

    cnst zMasked = iif(fMask, zOut, fTr.GetZ());

    cnst h = (zMasked - fTr.GetZ());

    cnst stepDz[5] = {0., 0., h * fvec(0.5), h * fvec(0.5), h};

    fvec f[5][7]    = {{0.}};    // ( d*/dz  ) [step]
    fvec F[5][7][7] = {{{0.}}};  // ( d *new [step] / d *old  )

    //   Runge-Kutta steps
    //

    fvec r0[7]    = {fTr.X(), fTr.Y(), fTr.Tx(), fTr.Ty(), fQp0, fTr.Time(), fTr.Vi()};
    fvec R0[7][7] = {{0.}};
    for (int i = 0; i < 7; ++i) {
      R0[i][i] = 1.;
    }

    for (int step = 1; step <= 4; ++step) {

      fvec rstep[7] = {{0.}};
      for (int i = 0; i < 7; ++i) {
        rstep[i] = r0[i] + stepDz[step] * f[step - 1][i];
      }
      fvec z = fTr.GetZ() + stepDz[step];

      ca::FieldValue B = Field.Get(rstep[0], rstep[1], z);
      fvec Bx          = B.x;
      fvec By          = B.y;
      fvec Bz          = B.z;

      fvec tx    = rstep[2];
      fvec ty    = rstep[3];
      fvec tx2   = tx * tx;
      fvec ty2   = ty * ty;
      fvec txty  = tx * ty;
      fvec L2    = fvec(1.) + tx2 + ty2;
      fvec L2i   = fvec(1.) / L2;
      fvec L     = sqrt(L2);
      fvec cL    = c_light * L;
      fvec cLqp0 = cL * fQp0;

      f[step][0]    = tx;
      F[step][0][2] = 1.;

      f[step][1]    = ty;
      F[step][1][3] = 1.;

      fvec f2tmp = txty * Bx - (fvec(1.) + tx2) * By + ty * Bz;
      f[step][2] = cLqp0 * f2tmp;

      F[step][2][2] = cLqp0 * (tx * f2tmp * L2i + ty * Bx - fvec(2.) * tx * By);
      F[step][2][3] = cLqp0 * (ty * f2tmp * L2i + tx * Bx + Bz);
      F[step][2][4] = cL * f2tmp;

      fvec f3tmp    = -txty * By - tx * Bz + (fvec(1.) + ty2) * Bx;
      f[step][3]    = cLqp0 * f3tmp;
      F[step][3][2] = cLqp0 * (tx * f3tmp * L2i - ty * By - Bz);
      F[step][3][3] = cLqp0 * (ty * f3tmp * L2i + fvec(2.) * ty * Bx - tx * By);
      F[step][3][4] = cL * f3tmp;

      f[step][4] = 0.;

      if (fDoFitVelocity) {
        fvec vi       = rstep[6];
        f[step][5]    = vi * L;
        F[step][5][2] = vi * tx / L;
        F[step][5][3] = vi * ty / L;
        F[step][5][4] = 0.;
        F[step][5][5] = 0.;
        F[step][5][6] = L;
      }
      else {
        fvec vi       = sqrt(fvec(1.) + fMass2 * fQp0 * fQp0) * fvec(constants::phys::SpeedOfLightInv);
        f[step][5]    = vi * L;
        F[step][5][2] = vi * tx / L;
        F[step][5][3] = vi * ty / L;
        F[step][5][4] =
          fMass2 * fQp0 * L / sqrt(fvec(1.) + fMass2 * fQp0 * fQp0) * fvec(constants::phys::SpeedOfLightInv);
        F[step][5][5] = 0.;
        F[step][5][6] = 0.;
      }

      f[step][6] = 0.;

    }  // end of Runge-Kutta step

    fvec r[7]    = {{0.}};  // extrapolated parameters
    fvec R[7][7] = {{0.}};  // Jacobian of the extrapolation

    cnst stepW[5] = {0., h / fvec(6.), h / fvec(3.), h / fvec(3.), h / fvec(6.)};

    fvec k[5][7][7] = {{0.}};
    for (int step = 1; step <= 4; ++step) {
      for (int i = 0; i < 7; i++) {
        for (int j = 0; j < 7; j++) {
          k[step][i][j] = F[step][i][j];
          for (int m = 0; m < 7; m++) {
            k[step][i][j] += stepDz[step] * F[step][i][m] * k[step - 1][m][j];
          }
        }
      }
    }

    for (int i = 0; i < 7; i++) {
      for (int j = 0; j < 7; j++) {
        R[i][j] = R0[i][j];
        for (int step = 1; step <= 4; step++) {
          R[i][j] += stepW[step] * k[step][i][j];
        }
      }
    }

    fvec dqp = fTr.Qp() - fQp0;

    for (int i = 0; i < 7; i++) {
      r[i] = r0[i];
      for (int step = 1; step <= 4; step++) {
        r[i] += stepW[step] * f[step][i];
      }
      // take into account linearisation at fQp0
      r[i] += R[i][4] * dqp;
    }

    // update parameters

    fTr.X()    = r[0];
    fTr.Y()    = r[1];
    fTr.Tx()   = r[2];
    fTr.Ty()   = r[3];
    fTr.Qp()   = r[4];
    fTr.Time() = r[5];
    fTr.Vi()   = r[6];

    //fTr.Vi()( fTr.Vi() < fvec(TrackParamV::kClightNsInv) ) = fvec(TrackParamV::kClightNsInv);
    fTr.Z() = zMasked;

    //          covariance matrix transport

    fvec C[7][7];
    for (int i = 0; i < 7; i++) {
      for (int j = 0; j < 7; j++) {
        C[i][j] = fTr.C(i, j);
      }
    }

    fvec RC[7][7];
    for (int i = 0; i < 7; i++) {
      for (int j = 0; j < 7; j++) {
        RC[i][j] = 0.;
        for (int m = 0; m < 7; m++) {
          RC[i][j] += R[i][m] * C[m][j];
        }
      }
    }
    for (int i = 0; i < 7; i++) {
      for (int j = 0; j < 7; j++) {
        fvec Cij = 0.;
        for (int m = 0; m < 7; m++) {
          Cij += RC[i][m] * R[j][m];
        }
        fTr.C(i, j) = Cij;
      }
    }
  }


  void TrackFit::ExtrapolateLine(fvec z_out, const ca::FieldRegion<fvec>& F)
  {
    // extrapolate the track assuming fQp0 == 0
    // TODO: write special simplified procedure
    //
    auto qp0 = fQp0;
    fQp0     = fvec(0.);
    Extrapolate(z_out, F);
    fQp0 = qp0;
  }

  void TrackFit::ExtrapolateLineNoField(fvec zOut)
  {
    // extrapolate the track assuming no field

    //   x += dz * tx
    //   y += dz * ty
    //   t += dz * sqrt ( 1 + tx*tx + ty*ty )  * vi

    if (0) {  // debug: full Runge-Kutta extrapolation with zero field values
      FieldRegion<fvec> F;
      ExtrapolateStep(zOut, F);
      return;
    }

    auto& t = fTr;  // use reference to shorten the text

    cnst zMasked = iif(fMask, zOut, t.GetZ());

    cnst dz = (zMasked - t.GetZ());

    fvec tx = t.GetTx();
    fvec ty = t.GetTy();
    fvec vi = t.GetVi();

    fvec L = sqrt(fvec(1.) + tx * tx + ty * ty);

    fvec j52 = dz * tx * vi / L;
    fvec j53 = dz * ty * vi / L;
    fvec j56 = dz * L;

    // transport parameters

    t.X() += tx * dz;
    t.Y() += ty * dz;
    t.Time() += L * vi * dz;
    t.Z() = zMasked;

    // transport covariance matrix

    //     ( 1   0  dz   0   0   0   0 )
    //     ( 0   1   0  dz   0   0   0 )
    //     ( 0   0   1   0   0   0   0 )
    // J = ( 0   0   0   1   0   0   0 )
    //     ( 0   0   0   0   1   0   0 )
    //     ( 0   0 j52 j53   0   1 j56 )
    //     ( 0   0   0   0   0   0   1 )


    // JC = J * C

    fvec jc00 = t.C00() + dz * t.C20();
    //fvec jc01 = t.C01() + dz * t.C21();
    fvec jc02 = t.C02() + dz * t.C22();
    //fvec jc03 = t.C03() + dz * t.C23();
    //fvec jc04 = t.C04() + dz * t.C24();
    //fvec jc05 = t.C05() + dz * t.C25();
    //fvec jc06 = t.C06() + dz * t.C26();


    fvec jc10 = t.C10() + dz * t.C30();
    fvec jc11 = t.C11() + dz * t.C31();
    fvec jc12 = t.C12() + dz * t.C32();
    fvec jc13 = t.C13() + dz * t.C33();
    //fvec jc14 = t.C14() + dz * t.C34();
    //fvec jc15 = t.C15() + dz * t.C35();
    //fvec jc16 = t.C16() + dz * t.C36();

    // jc2? = t.C2?
    // jc3? = t.C3?
    // jc4? = t.C4?

    fvec jc50 = t.C50() + j52 * t.C20() + j53 * t.C30() + j56 * t.C60();
    fvec jc51 = t.C51() + j52 * t.C21() + j53 * t.C31() + j56 * t.C61();
    fvec jc52 = t.C52() + j52 * t.C22() + j53 * t.C32() + j56 * t.C62();
    fvec jc53 = t.C53() + j52 * t.C23() + j53 * t.C33() + j56 * t.C63();
    fvec jc54 = t.C54() + j52 * t.C24() + j53 * t.C34() + j56 * t.C64();
    fvec jc55 = t.C55() + j52 * t.C25() + j53 * t.C35() + j56 * t.C65();
    fvec jc56 = t.C56() + j52 * t.C26() + j53 * t.C36() + j56 * t.C66();

    // jc6? = t.C6?

    // transpose J
    //
    //      (  1   0   0   0   0   0   0 )
    //      (  0   1   0   0   0   0   0 )
    //      ( dz   0   1   0   0 j52   0 )
    // J' = (  0  dz   0   1   0 j53   0 )
    //      (  0   0   0   0   1   0   0 )
    //      (  0   0   0   0   0   1   0 )
    //      (  0   0   0   0   0 j56   1 )


    // C = JC * J'

    t.C00() = jc00 + jc02 * dz;
    t.C10() = jc10 + jc12 * dz;
    t.C20() = t.C20() + t.C22() * dz;
    t.C30() = t.C30() + t.C32() * dz;
    t.C40() = t.C40() + t.C42() * dz;
    t.C50() = jc50 + jc52 * dz;
    t.C60() = t.C60() + t.C62() * dz;

    t.C11() = jc11 + jc13 * dz;
    t.C21() = t.C21() + t.C23() * dz;
    t.C31() = t.C31() + t.C33() * dz;
    t.C41() = t.C41() + t.C43() * dz;
    t.C51() = jc51 + jc53 * dz;
    t.C61() = t.C61() + t.C63() * dz;

    // t.C22 = jc22 == t.C22 -> unchanged
    // t.C32 = jc32 == t.C32 -> unchanged
    // t.C42 = jc42 == t.C42 -> unchanged
    t.C52() = jc52;
    // t.C62 = jc62 == t.C62 -> unchanged

    // t.C33 = jc33 == t.C33 -> unchanged
    // t.C43 = jc43 == t.C43 -> unchanged
    t.C53() = jc53;
    // t.C63 = jc63 == t.C63 -> unchanged

    // t.C44 = jc44 == t.C44 -> unchanged
    t.C54() = jc54;
    // t.C64 = jc64 == t.C64 -> unchanged

    t.C55() = jc55 + jc52 * j52 + jc53 * j53 + jc56 * j56;
    t.C65() = t.C65() + t.C62() * j52 + t.C63() * j53 + t.C66() * j56;

    // t.C66 = jc66 = t.C66 -> unchanged
  }


  void TrackFit::MultipleScattering(fvec radThick, fvec qp0)
  {
    cnst ONE = 1.;

    fvec tx    = fTr.Tx();
    fvec ty    = fTr.Ty();
    fvec txtx  = tx * tx;
    fvec tyty  = ty * ty;
    fvec txtx1 = txtx + ONE;
    fvec h     = txtx + tyty;
    fvec t     = sqrt(txtx1 + tyty);
    fvec h2    = h * h;
    fvec qp0t  = qp0 * t;

    cnst c1 = 0.0136f, c2 = c1 * 0.038f, c3 = c2 * 0.5f, c4 = -c3 / 2.0f, c5 = c3 / 3.0f, c6 = -c3 / 4.0f;

    fvec s0 = (c1 + c2 * log(radThick) + c3 * h + h2 * (c4 + c5 * h + c6 * h2)) * qp0t;
    //fvec a = ( (ONE+mass2*qp0*qp0t)*radThick*s0*s0 );
    fvec a = ((t + fMass2 * qp0 * qp0t) * radThick * s0 * s0);

    fTr.C22()(fMask) += txtx1 * a;
    fTr.C32()(fMask) += tx * ty * a;
    fTr.C33()(fMask) += (ONE + tyty) * a;
  }

  void TrackFit::MultipleScatteringInThickMaterial(fvec radThick, fvec thickness, bool fDownstream)
  {
    cnst ONE = 1.;

    fvec tx    = fTr.Tx();
    fvec ty    = fTr.Ty();
    fvec txtx  = tx * tx;
    fvec tyty  = ty * ty;
    fvec txtx1 = txtx + ONE;
    fvec h     = txtx + tyty;
    fvec t     = sqrt(txtx1 + tyty);
    fvec h2    = h * h;
    fvec qp0t  = fQp0 * t;

    cnst c1(0.0136), c2 = c1 * fvec(0.038), c3 = c2 * fvec(0.5), c4 = -c3 / fvec(2.0), c5 = c3 / fvec(3.0),
                     c6 = -c3 / fvec(4.0);

    fvec s0 = (c1 + c2 * log(radThick) + c3 * h + h2 * (c4 + c5 * h + c6 * h2)) * qp0t;
    //fvec a = ( (ONE+mass2*qp0*qp0t)*radThick*s0*s0 );
    fvec a = ((t + fMass2 * fQp0 * qp0t) * radThick * s0 * s0);

    fvec D   = (fDownstream) ? fvec(1.) : fvec(-1.);
    fvec T23 = (thickness * thickness) / fvec(3.0);
    fvec T2  = thickness / fvec(2.0);

    fTr.C00()(fMask) += txtx1 * a * T23;
    fTr.C10()(fMask) += tx * ty * a * T23;
    fTr.C20()(fMask) += txtx1 * a * D * T2;
    fTr.C30()(fMask) += tx * ty * a * D * T2;

    fTr.C11()(fMask) += (ONE + tyty) * a * T23;
    fTr.C21()(fMask) += tx * ty * a * D * T2;
    fTr.C31()(fMask) += (ONE + tyty) * a * D * T2;

    fTr.C22()(fMask) += txtx1 * a;
    fTr.C32()(fMask) += tx * ty * a;
    fTr.C33()(fMask) += (ONE + tyty) * a;
  }


  void TrackFit::EnergyLossCorrection(fvec radThick, fvec upstreamDirection)
  {
    cnst qp2cut(1. / (10. * 10.));  // 10 GeV cut
    cnst qp02 = max(fQp0 * fQp0, qp2cut);
    cnst p2   = fvec(1.) / qp02;
    cnst E2   = fMass2 + p2;

    cnst bethe = ApproximateBetheBloch(p2 / fMass2);

    fvec tr = sqrt(fvec(1.f) + fTr.Tx() * fTr.Tx() + fTr.Ty() * fTr.Ty());

    cnst dE = bethe * radThick * tr * 2.33f * 9.34961f;

    cnst ECorrected  = sqrt(E2) + upstreamDirection * dE;
    cnst E2Corrected = ECorrected * ECorrected;

    fvec corr = sqrt(p2 / (E2Corrected - fMass2));
    fmask ok  = (corr == corr) && fMask;
    corr      = iif(ok, corr, fvec::One());

    fQp0 *= corr;
    fTr.Qp() *= corr;
    fTr.C40() *= corr;
    fTr.C41() *= corr;
    fTr.C42() *= corr;
    fTr.C43() *= corr;
    fTr.C44() *= corr * corr;
    fTr.C54() *= corr;
  }


  void TrackFit::EnergyLossCorrection(int atomicZ, float atomicA, float rho, float radLen, fvec radThick,
                                      fvec upstreamDirection)
  {
    cnst qp2cut(1. / (10. * 10.));  // 10 GeV cut
    cnst qp02 = max(fQp0 * fQp0, qp2cut);
    cnst p2   = fvec(1.) / qp02;
    cnst E2   = fMass2 + p2;

    fvec i;
    if (atomicZ < 13)
      i = (12. * atomicZ + 7.) * 1.e-9;
    else
      i = (9.76 * atomicZ + 58.8 * std::pow(atomicZ, -0.19)) * 1.e-9;

    cnst bethe = ApproximateBetheBloch(p2 / fMass2, rho, 0.20, 3.00, i, atomicZ / atomicA);

    fvec tr = sqrt(fvec(1.f) + fTr.Tx() * fTr.Tx() + fTr.Ty() * fTr.Ty());

    fvec dE = bethe * radThick * tr * radLen * rho;

    cnst E2Corrected = (sqrt(E2) + upstreamDirection * dE) * (sqrt(E2) + upstreamDirection * dE);
    fvec corr        = sqrt(p2 / (E2Corrected - fMass2));
    fmask ok         = (corr == corr) && fMask;
    corr             = iif(ok, corr, fvec::One());

    fQp0 *= corr;
    fTr.Qp() *= corr;

    fvec P(sqrt(p2));  // GeV

    fvec Z(atomicZ);
    fvec A(atomicA);
    fvec RHO(rho);

    fvec STEP = radThick * tr * radLen;
    fvec EMASS(0.511 * 1e-3);  // GeV

    fvec BETA  = P / sqrt(E2Corrected);
    fvec GAMMA = sqrt(E2Corrected) / fMass;

    // Calculate xi factor (KeV).
    fvec XI = (fvec(153.5) * Z * STEP * RHO) / (A * BETA * BETA);

    // Maximum energy transfer to atomic electron (KeV).
    fvec ETA   = BETA * GAMMA;
    fvec ETASQ = ETA * ETA;
    fvec RATIO = EMASS / fMass;
    fvec F1    = fvec(2.) * EMASS * ETASQ;
    fvec F2    = fvec(1.) + fvec(2.) * RATIO * GAMMA + RATIO * RATIO;
    fvec EMAX  = fvec(1e6) * F1 / F2;

    fvec DEDX2 = XI * EMAX * (fvec(1.) - (BETA * BETA / fvec(2.))) * fvec(1e-12);

    fvec P2    = P * P;
    fvec SDEDX = (E2 * DEDX2) / (P2 * P2 * P2);

    //   T.fTr.C40() *= corr;
    //   T.fTr.C41() *= corr;
    //   T.fTr.C42() *= corr;
    //   T.fTr.C43() *= corr;
    // T.fTr.C44() *= corr*corr;
    fTr.C44() += abs(SDEDX);
  }


  void TrackFit::GetExtrapolatedXYline(fvec z, const ca::FieldRegion<fvec>& F, fvec& extrX, fvec& extrY,
                                       std::array<fvec, ca::TrackParamV::kNtrackParam>& Jx,
                                       std::array<fvec, ca::TrackParamV::kNtrackParam>& Jy) const
  {
    // extrapolate track assuming it is straight (qp==0)
    // return the extrapolated X, Y and the derivatives of the extrapolated X and Y

    cnst c_light(0.000299792458), c1(1.), c2i(0.5), c6i(1. / 6.), c12i(1. / 12.);

    cnst tx = fTr.GetTx();
    cnst ty = fTr.GetTy();
    fvec dz = z - fTr.GetZ();

    fvec dz2     = dz * dz;
    fvec dzc6i   = dz * c6i;
    fvec dz2c12i = dz2 * c12i;

    fvec xx = tx * tx;
    fvec yy = ty * ty;
    fvec xy = tx * ty;

    fvec Ay = -xx - c1;
    fvec Bx = yy + c1;

    fvec ctdz2 = c_light * sqrt(c1 + xx + yy) * dz2;

    fvec Sx = F.cx0 * c2i + F.cx1 * dzc6i + F.cx2 * dz2c12i;
    fvec Sy = F.cy0 * c2i + F.cy1 * dzc6i + F.cy2 * dz2c12i;
    fvec Sz = F.cz0 * c2i + F.cz1 * dzc6i + F.cz2 * dz2c12i;

    extrX = fTr.GetX() + tx * dz;
    extrY = fTr.GetY() + ty * dz;

    Jx.fill(fvec(0.));  // for a case
    Jy.fill(fvec(0.));

    Jx[0] = fvec(1.);
    Jx[1] = fvec(0.);
    Jx[2] = dz;
    Jx[3] = fvec(0.);
    Jx[4] = ctdz2 * (Sx * xy + Sy * Ay + Sz * ty);
    Jx[5] = fvec(0.);
    Jx[6] = fvec(0.);

    Jy[0] = fvec(0.);
    Jy[1] = fvec(1.);
    Jy[2] = fvec(0.);
    Jy[3] = dz;
    Jy[4] = ctdz2 * (Sx * Bx - Sy * xy - Sz * tx);
    Jy[5] = fvec(0.);
    Jx[6] = fvec(0.);
  }


  void TrackFit::FilterWithTargetAtLine(fvec targZ, const ca::MeasurementXy<fvec>& targXY,
                                        const ca::FieldRegion<fvec>& F)
  {
    // Add the target constraint to a straight line track

    fvec eX, eY;
    std::array<fvec, ca::TrackParamV::kNtrackParam> Jx, Jy;
    GetExtrapolatedXYline(targZ, F, eX, eY, Jx, Jy);
    FilterExtrapolatedXY(targXY, eX, eY, Jx, Jy);
  }

  fvec TrackFit::ApproximateBetheBloch(fvec bg2)
  {
    //
    // This is the parameterization of the Bethe-Bloch formula inspired by Geant.
    //
    // bg2  - (beta*gamma)^2
    // kp0 - density [g/cm^3]
    // kp1 - density effect first junction point
    // kp2 - density effect second junction point
    // kp3 - mean excitation energy [GeV]
    // kp4 - mean Z/A
    //
    // The default values for the kp* parameters are for silicon.
    // The returned value is in [GeV/(g/cm^2)].
    //

    cnst kp0 = 2.33f;
    cnst kp1 = 0.20f;
    cnst kp2 = 3.00f;
    cnst kp3 = 173e-9f;
    cnst kp4 = 0.49848f;

    constexpr float mK   = 0.307075e-3f;  // [GeV*cm^2/g]
    constexpr float _2me = 1.022e-3f;     // [GeV/c^2]
    cnst rho             = kp0;
    cnst x0              = kp1 * 2.303f;
    cnst x1              = kp2 * 2.303f;
    cnst mI              = kp3;
    cnst mZA             = kp4;
    cnst maxT            = _2me * bg2;  // neglecting the electron mass

    //*** Density effect
    fvec d2(0.f);
    cnst x    = 0.5f * log(bg2);
    cnst lhwI = log(28.816f * 1e-9f * sqrt(rho * mZA) / mI);

    fmask init = x > x1;
    d2         = iif(init, lhwI + x - 0.5f, fvec::Zero());
    cnst r     = (x1 - x) / (x1 - x0);
    init       = (x > x0) & (x1 > x);
    d2         = iif(init, lhwI + x - 0.5f + (0.5f - lhwI - x0) * r * r * r, d2);

    return mK * mZA * (fvec(1.f) + bg2) / bg2
           * (0.5f * log(_2me * bg2 * maxT / (mI * mI)) - bg2 / (fvec(1.f) + bg2) - d2);
  }

  fvec TrackFit::ApproximateBetheBloch(fvec bg2, fvec kp0, fvec kp1, fvec kp2, fvec kp3, fvec kp4)
  {
    //
    // This is the parameterization of the Bethe-Bloch formula inspired by Geant.
    //
    // bg2  - (beta*gamma)^2
    // kp0 - density [g/cm^3]
    // kp1 - density effect first junction point
    // kp2 - density effect second junction point
    // kp3 - mean excitation energy [GeV]
    // kp4 - mean Z/A
    //
    // The default values for the kp* parameters are for silicon.
    // The returned value is in [GeV/(g/cm^2)].
    //

    //   cnst &kp0 = 2.33f;
    //   cnst &kp1 = 0.20f;
    //   cnst &kp2 = 3.00f;
    //   cnst &kp3 = 173e-9f;
    //   cnst &kp4 = 0.49848f;

    constexpr float mK   = 0.307075e-3f;  // [GeV*cm^2/g]
    constexpr float _2me = 1.022e-3f;     // [GeV/c^2]
    fvec rho             = kp0;
    cnst x0              = kp1 * 2.303f;
    cnst x1              = kp2 * 2.303f;
    fvec mI              = kp3;
    fvec mZA             = kp4;
    cnst maxT            = _2me * bg2;  // neglecting the electron mass

    //*** Density effect
    fvec d2(0.f);
    cnst x    = 0.5f * log(bg2);
    cnst lhwI = log(28.816f * 1e-9f * sqrt(rho * mZA) / mI);

    fmask init = x > x1;
    d2         = iif(init, lhwI + x - 0.5f, fvec::Zero());
    cnst r     = (x1 - x) / (x1 - x0);
    init       = (x > x0) & (x1 > x);
    d2         = iif(init, lhwI + x - 0.5f + (0.5f - lhwI - x0) * r * r * r, d2);

    return mK * mZA * (fvec(1.f) + bg2) / bg2
           * (0.5f * log(_2me * bg2 * maxT / (mI * mI)) - bg2 / (fvec(1.f) + bg2) - d2);
  }


  std::tuple<fvec, fvec> TrackFit::GetChi2XChi2U(ca::MeasurementXy<fvec> m, fvec x, fvec y, fvec C00, fvec C10,
                                                 fvec C11)
  {

    fvec chi2x{0.};

    {  // filter X measurement
      fvec zeta = x - m.X();

      // F = CH'
      fvec F0 = C00;
      fvec F1 = C10;

      fvec HCH = F0;

      fvec wi     = fvec(1.) / (m.Dx2() + HCH);
      fvec zetawi = zeta * wi;
      chi2x       = m.NdfX() * zeta * zetawi;

      fvec K1 = F1 * wi;

      x -= F0 * zetawi;
      y -= F1 * zetawi;

      C00 -= F0 * F0 * wi;
      C10 -= K1 * F0;
      C11 -= K1 * F1;
    }

    fvec chi2u{0.};

    {  // filter U measurement, we need only chi2 here
      fvec cosPhi = -m.Dxy() / m.Dx2();
      fvec u      = cosPhi * m.X() + m.Y();
      fvec du2    = m.Dy2() + cosPhi * m.Dxy();

      fvec zeta = cosPhi * x + y - u;

      // F = CH'
      fvec F0 = cosPhi * C00 + C10;
      fvec F1 = cosPhi * C10 + C11;

      fvec HCH = (F0 * cosPhi + F1);

      chi2u += m.NdfY() * zeta * zeta / (du2 + HCH);
    }

    return std::tuple<fvec, fvec>(chi2x, chi2u);
  }


  void TrackFit::GuessTrack(const fvec& trackZ, const fvec hitX[], const fvec hitY[], const fvec hitZ[],
                            const fvec hitT[], const fvec By[], const fmask hitW[], const fmask hitWtime[], int NHits)
  {
    // gives nice initial approximation for x,y,tx,ty - almost same as KF fit. qp - is shifted by 4%, resid_ual - ~3.5% (KF fit resid_ual - 1%).

    const fvec c_light(0.000299792458), c_light_i(fvec(1.) / c_light);

    fvec A0 = 0., A1 = 0., A2 = 0., A3 = 0., A4 = 0., A5 = 0.;
    fvec a0 = 0., a1 = 0., a2 = 0.;
    fvec b0 = 0., b1 = 0., b2 = 0.;

    fvec time       = 0.;
    fmask isTimeSet = fmask::Zero();

    fvec prevZ = 0.;
    fvec sy = 0., Sy = 0.;  // field integrals

    for (int i = 0; i < NHits; i++) {

      fvec w = iif(hitW[i], fvec::One(), fvec::Zero());

      fmask setTime = (!isTimeSet) && hitWtime[i] && hitW[i];
      time(setTime) = hitT[i];
      isTimeSet     = isTimeSet || setTime;

      fvec x = hitX[i];
      fvec y = hitY[i];
      fvec z = hitZ[i] - trackZ;

      {
        fvec dZ = z - prevZ;
        Sy(hitW[i]) += dZ * sy + fvec(0.5) * dZ * dZ * By[i];
        sy(hitW[i]) += dZ * By[i];
        prevZ(hitW[i]) = z;
      }

      fvec S = Sy;

      fvec wz = w * z;
      fvec wS = w * S;

      A0 += w;
      A1 += wz;
      A2 += wz * z;
      A3 += wS;
      A4 += wS * z;
      A5 += wS * S;

      a0 += w * x;
      a1 += wz * x;
      a2 += wS * x;

      b0 += w * y;
      b1 += wz * y;
      b2 += wS * y;
    }

    fvec m00 = A0;
    fvec m01 = A1;
    fvec m02 = A3;

    fvec m11 = A2;
    fvec m12 = A4;

    fvec m22 = A5;

    fvec m21 = m12;

    // { m00 m01 m02 }       ( a0 )
    // { m01 m11 m12 } = x * ( a1 )
    // { m02 m21 m22 }       ( a2 )

    {  // triangulation row 0
      m11 = m00 * m11 - m01 * m01;
      m12 = m00 * m12 - m01 * m02;
      a1  = m00 * a1 - m01 * a0;

      m21 = m00 * m21 - m02 * m01;
      m22 = m00 * m22 - m02 * m02;
      a2  = m00 * a2 - m02 * a0;
    }

    {  // triangulation step row 1
      m22 = m11 * m22 - m21 * m12;
      a2  = m11 * a2 - m21 * a1;
    }

    fvec L = 0.;
    {  // diagonalization row 2
      L(abs(m22) > fvec(1.e-4)) = a2 / m22;
      a1 -= L * m12;
      a0 -= L * m02;
    }

    {  // diagonalization row 1
      fTr.Tx() = a1 / m11;
      a0 -= fTr.Tx() * m01;
    }

    {  // diagonalization row 0
      fTr.X() = a0 / m00;
    }

    fvec txtx1 = fvec(1.) + fTr.Tx() * fTr.Tx();
    L          = L / txtx1;
    fvec L1    = L * fTr.Tx();

    A1 = A1 + A3 * L1;
    A2 = A2 + (A4 + A4 + A5 * L1) * L1;
    b1 += b2 * L1;

    // { A0 A1 } = x * ( b0 )
    // { A1 A2 }       ( b1 )

    A2 = A0 * A2 - A1 * A1;
    b1 = A0 * b1 - A1 * b0;

    fTr.Ty() = b1 / A2;
    fTr.Y()  = (b0 - A1 * fTr.Ty()) / A0;

    fTr.Qp()   = -L * c_light_i / sqrt(txtx1 + fTr.Ty() * fTr.Ty());
    fTr.Time() = time;
    fTr.Z()    = trackZ;
    fTr.Vi()   = constants::phys::SpeedOfLightInv;
    fQp0       = fTr.Qp();
  }

}  // namespace cbm::algo::ca
