/* Copyright (C) 2023-2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CaQaBuilder.cxx
/// \date   20.11.2023
/// \brief  A QA module for CA tracking (implementation)
/// \author S.Zharko <s.zharko@gsi.de>

#include "CaQaBuilder.h"

#include "CaConstants.h"
#include "CaInputData.h"
#include "CaParameters.h"
#include "CaTrack.h"

#include <fmt/format.h>

using cbm::algo::CanvasConfig;
using cbm::algo::Histo1D;
using cbm::algo::PadConfig;
using cbm::algo::QaData;
using cbm::algo::ca::constants::math::Pi;
using cbm::algo::ca::qa::Builder;


// ---------------------------------------------------------------------------------------------------------------------
//
void Builder::Init()
{
  using fmt::format;

  if (!fpSender.get()) {
    return;
  }

  // ---- Init histograms
  // TODO: Provide definition from config
  constexpr int nParPads                              = 4;
  std::array<std::string, knTrkParPoints> vsPointName = {"first", "last"};
  for (int i = 0; i < knTrkParPoints; ++i) {
    {
      auto sName      = format("track_{}_theta", vsPointName[i]);
      auto sTitl      = format("#theta at {} hit; #theta", vsPointName[i]);
      fvphTrkTheta[i] = fQaData.MakeObj<Histo1D>(62, 0., 90., sName, sTitl);
    }
    {
      auto sName    = format("track_{}_phi", vsPointName[i]);
      auto sTitl    = format("#phi at {} hit; #phi", vsPointName[i]);
      fvphTrkPhi[i] = fQaData.MakeObj<Histo1D>(62, -180., 180., sName, sTitl);
    }
    {
      auto sName        = format("track_{}_chi2_ndf", vsPointName[i]);
      auto sTitl        = format("#chi^{{2}}/NDF at {} hit; #chi^{{2}}/NDF", vsPointName[i]);
      fvphTrkChi2Ndf[i] = fQaData.MakeObj<Histo1D>(100, 0., 20., sName, sTitl);
    }
    {
      auto sName    = format("track_{}_hit_station", vsPointName[i]);
      auto sTitl    = format("Station of {} hit;ID_{{sta}}", vsPointName[i]);
      fvphHitSta[i] = fQaData.MakeObj<Histo1D>(100, -.5, 10.5, sName, sTitl);
    }
  }
  fphTrkNofHits = fQaData.MakeObj<Histo1D>(11, -0.5, 10.5, "n_hits", "Number of hits;N_{hit}");

  // ---- Init canvases
  {
    // Track parameters at first/last track hit
    for (int i = 0; i < knTrkParPoints; ++i) {
      auto sCanvName = fmt::format("ca_trk_{}_hit", vsPointName[i]);
      auto sCanvTitl = fmt::format("Track parameters at {} hit", vsPointName[i]);
      auto canv      = CanvasConfig(sCanvName, sCanvTitl);
      std::vector<PadConfig> pads;
      pads.reserve(nParPads);
      for (int iPad = 0; iPad < nParPads; ++iPad) {
        auto& pad = pads.emplace_back();
        pad.SetLog(false, true);
        pad.SetGrid(true, true);
      }
      pads[0].RegisterHistogram(fvphTrkTheta[i], "hist");
      pads[1].RegisterHistogram(fvphTrkPhi[i], "hist");
      pads[2].RegisterHistogram(fvphTrkChi2Ndf[i], "hist");
      pads[3].RegisterHistogram(fvphHitSta[i], "hist");
      for (auto& pad : pads) {
        canv.AddPadConfig(pad);
      }
      fQaData.AddCanvasConfig(canv);
    }

    // Number of hits in track
    {
      auto canv = CanvasConfig("ca_nhits_in_trk", "Number of hit in track");
      auto pad  = PadConfig();
      pad.SetGrid(true, true);
      pad.SetLog(false, true);
      pad.RegisterHistogram(fphTrkNofHits, "hist");
      canv.AddPadConfig(pad);
      fQaData.AddCanvasConfig(canv);
    }
  }

  fQaData.Init(fpSender);
}

// ---------------------------------------------------------------------------------------------------------------------
//
void Builder::Build()
{
  if (!fpSender.get()) {
    return;
  }

  if (!fpParameters) {
    LOG(error) << "cbm::algo::ca::qa::Builder::Build(): parameters object is undefined";
    return;
  }
  if (!fpInputData) {
    LOG(error) << "cbm::algo::ca::qa::Builder::Build(): input data object is undefined";
    return;
  }
  if (!fpvTracks) {
    LOG(error) << "cbm::algo::ca::qa::Builder::Build(): tracks vector is undefined";
    return;
  }
  if (!fpvRecoHits) {
    LOG(error) << "cbm::algo::ca::qa::Builder::Build(): reco hit indices vector is undefined";
    return;
  }

  // Fill track histograms
  {
    int trkFirstHit = 0;  // Index of hit in fpvRecoHits
    for (auto trkIt = fpvTracks->begin(); trkIt != fpvTracks->end(); ++trkIt) {
      auto& track = *trkIt;
      int nHits   = track.fNofHits;
      // Indices of hits in fpInputData->GetHits()
      int iFstHit = (*fpvRecoHits)[trkFirstHit];
      int iLstHit = (*fpvRecoHits)[trkFirstHit + nHits - 1];

      // Distributions in different track points
      for (int ip = 0; ip < knTrkParPoints; ++ip) {
        int iHit           = (ip == 0 ? iFstHit : iLstHit);
        const auto& trkPar = (ip == 0 ? track.fParFirst : track.fParLast);
        const auto& hit    = fpInputData->GetHit(iHit);
        fvphTrkTheta[ip]->Add(trkPar.GetTheta() * 180. / Pi);
        fvphTrkPhi[ip]->Add(trkPar.GetPhi() * 180. / Pi);
        fvphTrkChi2Ndf[ip]->Add(trkPar.GetChiSq() / trkPar.GetNdf());
        fvphHitSta[ip]->Add(hit.Station());
      }
      // Other distributions
      fphTrkNofHits->Add(nHits);
      trkFirstHit += nHits;
    }
  }

  fQaData.Send(fpSender);
}
