/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CaQaBuilder.h
/// \date   20.11.2023
/// \brief  A QA module for CA tracking (header)
/// \author S.Zharko <s.zharko@gsi.de>

#pragma once

#include "CaHit.h"  // for HitIndex_t
#include "CaTimesliceHeader.h"
#include "CaVector.h"
#include "QaData.h"

namespace cbm::algo
{
  class Histo1D;
  namespace ca
  {
    template<typename DataT>
    class Parameters;
    class InputData;
    class Track;
  }  // namespace ca
}  // namespace cbm::algo::ca

namespace cbm::algo::ca::qa
{
  /// \class cbm::algo::ca::qa::Builder
  /// \brief Builder class for the CA tracking QA (header)
  ///
  class Builder {
   public:
    using TrackV_t    = ca::Vector<ca::Track>;
    using HitIndexV_t = ca::Vector<std::vector<std::pair<uint32_t, uint32_t>>>;

    /// \brief Default destructor
    /// \param pSender  Pointer to the histogram sender
    Builder(std::shared_ptr<HistogramSender> pSender) : fpSender(pSender){};

    /// \brief Constructor from the configuration object
    /// \param config  QA configuration object
    Builder() = default;

    /// \brief Copy constructor
    Builder(const Builder&) = delete;

    /// \brief Move constructor
    Builder(Builder&&) = delete;

    /// \brief Destructor
    ~Builder() = default;

    /// \brief Copy assignment operator
    Builder& operator=(const Builder&) = delete;

    /// \brief Move assignment operator
    Builder& operator=(Builder&&) = delete;

    /// \brief QA execution function
    void Build();

    /// \brief Initializes the QA
    void Init();

    /// \brief Checks, if the histogram sender is defined
    bool IsSenderDefined() const { return static_cast<bool>(fpSender.get()); }

    /// \brief Registers tracking input data
    /// \note  Call per TS
    void RegisterInputData(const InputData* pInputData) { fpInputData = pInputData; }

    /// \brief Registers track vector
    /// \note  Call per TS
    void RegisterTracks(const Vector<Track>* pvTracks) { fpvTracks = pvTracks; }

    /// \brief Registers reco hits indices
    /// \note  Call per TS
    void RegisterRecoHitIndices(const Vector<HitIndex_t>* pvRecoHits) { fpvRecoHits = pvRecoHits; }

    /// \brief Registers tracking parameters object
    /// \note  Call per run
    void RegisterParameters(const Parameters<fvec>* pParameters) { fpParameters = pParameters; }

   private:
    QaData fQaData{"CaQa"};  ///< QA data

    std::shared_ptr<HistogramSender> fpSender = nullptr;  ///< Histogram sender
    const Parameters<fvec>* fpParameters      = nullptr;  ///< Pointer to tracking parameters
    const InputData* fpInputData              = nullptr;  ///< Pointer to input data
    const Vector<Track>* fpvTracks            = nullptr;  ///< Pointer to tracks vector
    const Vector<HitIndex_t>* fpvRecoHits     = nullptr;  ///< Pointer to reco hit indices

    // ----- List of the histograms ------------------------------------------------------------------------------------
    static constexpr int knTrkParPoints = 2;  ///< Number of track points to build par distributions

    std::array<Histo1D*, knTrkParPoints> fvphTrkTheta   = {{0}};    ///< hist: theta at first/last hit
    std::array<Histo1D*, knTrkParPoints> fvphTrkPhi     = {{0}};    ///< hist: phi at first/last hit
    std::array<Histo1D*, knTrkParPoints> fvphTrkChi2Ndf = {{0}};    ///< hist: chi2/NDF at first/last hit
    std::array<Histo1D*, knTrkParPoints> fvphHitSta     = {{0}};    ///< hist: station of first/last hit
    Histo1D* fphTrkNofHits                              = nullptr;  ///< hist: number of hits in track
  };
}  // namespace cbm::algo::ca::qa
