/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CaQaData.h
/// \date   20.11.2023
/// \brief  A data structure for the QA module of the CA tracking (header)
/// \author S.Zharko <s.zharko@gsi.de>

#pragma once

#include "CaQaConfig.h"
#include "CaQaDefinitions.h"
#include "Histo1D.h"

#include <string>
#include <vector>

namespace cbm::algo::ca::qa
{
  /// \class cbm::algo::ca::qa::Data
  /// \brief QA data structure for CA tracking
  class Data {
   public:
    /// \brief Default constructor
    Data() = delete;

    /// \brief Constructor from parameters
    /// \param config  QA configuration object
    Data(const Config& config);

    /// \brief Copy constructor
    Data(const Data&) = default;

    /// \brief Move constructor
    Data(Data&&) = default;

    /// \brief Copy assignment operator
    Data& operator=(const Data&) = default;

    /// \brief Move assignment operator
    Data& operator=(Data&&) = default;

    /// \brief Destructor
    ~Data() = default;

    /// \brief Accesses 1D-histogram
    const Histo1D& H1(H1Key key) const { return fvH1[static_cast<size_t>(key)]; }

    /// \brief Accesses 1D-histogram (mutable)
    Histo1D& H1(H1Key key) { return fvH1[static_cast<size_t>(key)]; }

    /// \brief Accesses the full histogram vector
    const std::vector<Histo1D>& VectorOfH1() const { return fvH1; }

    /// \brief String representation of the data status
    std::string ToString() const;

   private:
    std::vector<Histo1D> fvH1 = {};  ///< Vector of 1D-histograms, indexed by H1Key
  };

}  // namespace cbm::algo::ca::qa
