/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   QaData.cxx
/// \date   12.02.2024
/// \brief  A unified data-structure to handle QA objects for the online reconstruction (implementation)
/// \author Sergei Zharko <s.zharko@gsi.de>

#include "QaData.h"

using cbm::algo::QaData;

// ---------------------------------------------------------------------------------------------------------------------
//
void QaData::Init(std::shared_ptr<HistogramSender> histSender)
{
  if (histSender.get()) {
    // Forming a histogram config message
    std::vector<std::pair<std::string, std::string>> vHistCfgs;
    vHistCfgs.reserve(std::distance(fvH1.begin(), fvH1.end()));
    for (const auto& hist : fvH1) {
      vHistCfgs.emplace_back(hist.Name(), fsName);
    }

    // Forming a canvas config message
    std::vector<std::pair<std::string, std::string>> vCanvCfgs;
    vCanvCfgs.reserve(fvsCanvCfgs.size());
    for (const auto& canv : fvsCanvCfgs) {
      vCanvCfgs.emplace_back(std::make_pair(canv.substr(0, canv.find_first_of(';')), canv));
    }

    histSender->PrepareAndSendMsg(std::pair<uint32_t, uint32_t>(vHistCfgs.size(), vCanvCfgs.size()),
                                  zmq::send_flags::sndmore);
    for (const auto& cfg : vHistCfgs) {
      histSender->PrepareAndSendMsg(cfg, zmq::send_flags::sndmore);
    }
    for (const auto& cfg : vCanvCfgs) {
      histSender->PrepareAndSendMsg(cfg, zmq::send_flags::sndmore);
    }
    // Histograms serialization and emission to close multi-part message
    histSender->PrepareAndSendMsg(std::vector<Histo1D>{}, zmq::send_flags::none);
  }
}

// ---------------------------------------------------------------------------------------------------------------------
//
void QaData::Reset()
{
  // TODO: clear the vectors of other object types here as well
  for (auto& hist : fvH1) {
    hist.Clear();
  }
}

// ---------------------------------------------------------------------------------------------------------------------
//
void QaData::Send(std::shared_ptr<HistogramSender> histoSender)
{
  std::vector<Histo1D> vHisto1D{std::begin(fvH1), std::end(fvH1)};
  histoSender->PrepareAndSendMsg(vHisto1D, zmq::send_flags::none);
  L_(info) << fsName << ": Published " << vHisto1D.size() << " 1D-histograms";
  this->Reset();
}
