/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   QaData.cxx
/// \date   12.02.2024
/// \brief  A unified data-structure to handle QA objects for the online reconstruction
/// \author Sergei Zharko <s.zharko@gsi.de>

#pragma once

#include "CanvasConfig.h"
#include "Histo1D.h"
#include "HistogramSender.h"

#include <forward_list>
#include <log.hpp>
#include <memory>
#include <string_view>
#include <type_traits>
#include <vector>

namespace cbm::algo
{
  /// \class QaData
  /// \brief Class to handle QA-objects in the online reconstruction
  class QaData {
   public:
    /// \brief Constructor
    /// \param name  Name of the QA module (appears as the directory name in the output)
    QaData(std::string_view name) : fsName(name) {}

    /// \brief Copy constructor
    QaData(const QaData&) = default;

    /// \brief Move constructor
    QaData(QaData&&) = default;

    /// \brief Copy assignment operator
    QaData& operator=(const QaData&) = default;

    /// \brief Move assignment operator
    QaData& operator=(QaData&&) = default;

    /// \brief Destructor
    ~QaData() = default;

    /// \brief Adds a canvas to the canvas config list
    /// \param canvas  A CanvasConfig object
    void AddCanvasConfig(const CanvasConfig& canvas) { fvsCanvCfgs.push_back(canvas.ToString()); }

    /// \brief Creates a QA-object and returns the pointer to it
    template<class Obj, typename... Args>
    Obj* MakeObj(Args... args);

    /// \brief Gets module name
    std::string_view GetName() const { return fsName; }

    /// \brief Sends QA initialization information to the HistogramSender
    /// \param histoSender  A pointer to the histogram sender
    void Init(std::shared_ptr<HistogramSender> histoSender);

    /// \brief Resets the histograms
    void Reset();

    /// \brief Sends QA data to the HistogramSender
    /// \param histoSender  A pointer to the histogram sender
    /// \note  Calls this->Reset() after sending the message to the histogram server
    void Send(std::shared_ptr<HistogramSender> histoSender);

   private:
    std::string fsName;                         ///< Name of the QA module (used as a directory name)
    std::forward_list<Histo1D> fvH1      = {};  ///< List of 1D-histograms
    std::vector<std::string> fvsCanvCfgs = {};  ///< Vector of canvas configs
  };

  // -------------------------------------------------------------------------------------------------------------------
  //
  template<class Obj, typename... Args>
  Obj* QaData::MakeObj(Args... args)
  {
    if constexpr (std::is_same_v<Obj, cbm::algo::Histo1D>) {
      return &(fvH1.emplace_front(args...));
    }
    return nullptr;
  }
}  // namespace cbm::algo
