/* Copyright (C) 2022-2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergey Gorbunov, Sergei Zharko [committer] */

/// @file   CbmTrackingDetectorInterfaceBase.h
/// @brief  Base abstract class for tracking detector interface to L1 (implementation of Checker)
/// @since  31.05.2022
/// @author S.Zharko (s.zharko@gsi.de)

#include "CbmTrackingDetectorInterfaceBase.h"

#include <Logger.h>

#include <iomanip>
#include <limits>
#include <set>
#include <vector>

// ---------------------------------------------------------------------------------------------------------------------
//
bool CbmTrackingDetectorInterfaceBase::Check() const
{
  bool res = true;
  LOG(info) << ToString();
  std::stringstream msg;
  msg << "Errors in the detector interface initialization for " << this->GetDetectorName() << ":\n";

  // Number of stations
  if (this->GetNtrackingStations() < 1) {
    msg << "\t- Number of stations is less then 1(" << this->GetNtrackingStations() << ")";
    res = false && res;
  }
  else {
    // Station individual parameters check
    for (int iSt = 0; iSt < this->GetNtrackingStations(); ++iSt) {
      std::string prefix = std::string("\t- Station ") + std::to_string(iSt) + " has ";

      {  // Position along Z-axis
        double z0 = this->GetZmin(iSt);
        double z1 = this->GetZref(iSt);
        double z2 = this->GetZmax(iSt);
        if (!std::isfinite(z0) || !std::isfinite(z1) || !std::isfinite(z2) || !(z0 < z1 && z1 < z2)) {
          msg << prefix << " wrong Z position (" << z0 << " < " << z1 << " < " << z2 << " cm)\n";
          res = false && res;
        }
      }

      {  // Size along X-axis
        auto xMax = this->GetXmax(iSt);
        auto xMin = this->GetXmin(iSt);
        if (!(xMax > xMin) || std::isnan(xMax) || std::isnan(xMin)) {
          msg << prefix << " zero, negative or NaN X-size (xMin = " << xMin << ", xMax = " << xMax << " cm)\n";
          res = false && res;
        }
      }

      {  // Size along Y-axis
        auto yMax = this->GetYmax(iSt);
        auto yMin = this->GetYmin(iSt);
        if (!(yMax > yMin) || std::isnan(yMax) || std::isnan(yMin)) {
          msg << prefix << " zero, negative or NaN Y-size (yMin = " << yMin << ", yMax = " << yMax << " cm)\n";
          res = false && res;
        }
      }
    }

    {  // Position along beam axis
      std::vector<double> zPositions(this->GetNtrackingStations());
      for (int iSt = 0; iSt < this->GetNtrackingStations(); ++iSt) {
        zPositions[iSt] = this->GetZref(iSt);
      }
      std::set<double> zPositionSet(zPositions.begin(), zPositions.end());
      if (zPositions.size() != zPositionSet.size()) {
        msg << "\t- Some of stations have the same z position component:\n";
        for (size_t iSt = 0; iSt < zPositions.size(); ++iSt) {
          msg << "\t\tstation " << iSt << ", z = " << zPositions[iSt] << " cm\n";
        }
        res = false && res;
      }
    }
  }

  if (!res) {
    LOG(error) << msg.str()
               << "\033[4mErrors above mean that the CA tracking cannot be used with the current version of "
               << this->GetDetectorName() << " setup. Please, check if the " << this->GetDetectorName()
               << " setup parameters and the corresponding tracking detector interface are initialized properly\033[0m";
  }

  return res;
}

// ---------------------------------------------------------------------------------------------------------------------
//
std::string CbmTrackingDetectorInterfaceBase::ToString() const
{
  using std::setfill;
  using std::setw;
  std::stringstream table;
  table << std::boolalpha;
  table << "\nTracking detector interface: " << setw(5) << GetDetectorName() << '\n';
  table << setw(5) << "st.No" << ' ';
  table << setw(10) << "z_ref[cm]" << ' ';
  table << setw(10) << "z_min[cm]" << ' ';
  table << setw(10) << "z_max[cm]" << ' ';
  table << setw(10) << "x_min[cm]" << ' ';
  table << setw(10) << "x_max[cm]" << ' ';
  table << setw(10) << "y_min[cm]" << ' ';
  table << setw(10) << "y_max[cm]" << ' ';
  table << setw(10) << "time info" << ' ';
  table << '\n';
  for (int iSt = 0; iSt < GetNtrackingStations(); ++iSt) {
    table << setw(5) << iSt << ' ';
    table << setw(10) << GetZref(iSt) << ' ';
    table << setw(10) << GetZmin(iSt) << ' ';
    table << setw(10) << GetZmax(iSt) << ' ';
    table << setw(10) << GetXmin(iSt) << ' ';
    table << setw(10) << GetXmax(iSt) << ' ';
    table << setw(10) << GetYmin(iSt) << ' ';
    table << setw(10) << GetYmax(iSt) << ' ';
    table << setw(10) << IsTimeInfoProvided(iSt) << ' ';
    table << '\n';
  }
  return table.str();
}
