/* Copyright (C) 2022 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergey Gorbunov, Sergei Zharko [committer] */

/***************************************************************************************************
 * @file   CbmTofTrackingInterface.cxx
 * @brief  Input data and parameters interface from TOF subsystem used in L1 tracker (definition)
 * @since  23.06.2022
 * @author S.Zharko <s.zharko@gsi.de>
 ***************************************************************************************************/

#include "CbmTofTrackingInterface.h"

#include "CbmTofCreateDigiPar.h"

#include "FairDetector.h"
#include "FairRunAna.h"
#include <Logger.h>

#include <limits>

ClassImp(CbmTofTrackingInterface)

  CbmTofTrackingInterface* CbmTofTrackingInterface::fpInstance = nullptr;

//-------------------------------------------------------------------------------------------------------------------------------------
//
CbmTofTrackingInterface::CbmTofTrackingInterface() : FairTask("CbmTofTrackingInterface")
{
  if (!fpInstance) { fpInstance = this; }
}

//-------------------------------------------------------------------------------------------------------------------------------------
//
CbmTofTrackingInterface::~CbmTofTrackingInterface()
{
  if (fpInstance == this) { fpInstance = nullptr; }
}

//-------------------------------------------------------------------------------------------------------------------------------------
//
InitStatus CbmTofTrackingInterface::Init()
{
  // create digitization parameters from geometry file
  auto tofDigiPar = new CbmTofCreateDigiPar("TOF Digi Producer", "TOF task");
  LOG(info) << "Create DigiPar";
  tofDigiPar->Init();

  // ** ToF tracking station geometrical information initialization **

  // Init ToF stations position z-components. For each ToF tracking station the position z-component is calculated
  // as an average of the components for each ToF module inside the tracking station.
  std::vector<int> nTofStationModules(this->GetNtrackingStations(),
                                      0);  // Number of ToF modules for a given tracking station

  auto nStations = this->GetNtrackingStations();
  fTofStationXMin.clear();
  fTofStationXMax.clear();
  fTofStationYMin.clear();
  fTofStationYMax.clear();
  fTofStationZ.clear();
  fTofStationZMin.clear();
  fTofStationZMax.clear();
  fTofStationXMin.resize(nStations, std::numeric_limits<double>::max());
  fTofStationXMax.resize(nStations, std::numeric_limits<double>::lowest());
  fTofStationYMin.resize(nStations, std::numeric_limits<double>::max());
  fTofStationYMax.resize(nStations, std::numeric_limits<double>::lowest());
  fTofStationZ.resize(nStations, 0);
  fTofStationZMin.resize(nStations, std::numeric_limits<double>::max());
  fTofStationZMax.resize(nStations, std::numeric_limits<double>::lowest());

  fmAddressToStation.clear();
  for (int iSmType = 0; iSmType < fDigiBdfPar->GetNbSmTypes(); ++iSmType) {
    for (int iSm = 0; iSm < fDigiBdfPar->GetNbSm(iSmType); ++iSm) {
      for (int iRpc = 0; iRpc < fDigiBdfPar->GetNbRpc(iSmType); ++iRpc) {
        auto address = CbmTofAddress::GetUniqueAddress(iSm, iRpc, 0, 0, iSmType);
        int iStation = fDigiBdfPar->GetTrackingStation(iSmType, iSm, iRpc);  // Local index of tracking station
        fmAddressToStation[address] = iStation;

        auto* pChannelInfo = dynamic_cast<CbmTofCell*>(fDigiPar->GetCell(address));
        if (nullptr == pChannelInfo) {
          LOG(warn) << fName << ": CbmTofCell object is not defined for iSmType = " << iSmType << ", iSm = " << iSm
                    << ", iRpc = " << iRpc;
          continue;
        }

        // Tracking station sizes
        auto chPosX       = pChannelInfo->GetX();
        auto chPosY       = pChannelInfo->GetY();
        auto chPosZ       = pChannelInfo->GetZ();
        auto chSizeX      = pChannelInfo->GetSizex();
        auto chSizeY      = pChannelInfo->GetSizey();
        auto chLoBoarderX = chPosX - chSizeX;
        auto chUpBoarderX = chPosX + chSizeX;
        auto chLoBoarderY = chPosY - chSizeY;
        auto chUpBoarderY = chPosY + chSizeY;

        // Cuts on Bmon and undefined station ID
        if (5 == iSmType) { continue; }  // Skip Bmon
        if (iStation < 0) { continue; }

        fTofStationZ[iStation] += chPosZ;
        // FIXME: Sizes of stations are incorrect
        if (chPosZ > fTofStationZMax[iStation]) { fTofStationZMax[iStation] = chPosZ; }
        if (chPosZ < fTofStationZMin[iStation]) { fTofStationZMin[iStation] = chPosZ; }
        if (chUpBoarderX > fTofStationXMax[iStation]) { fTofStationXMax[iStation] = chUpBoarderX; }
        if (chUpBoarderY > fTofStationYMax[iStation]) { fTofStationYMax[iStation] = chUpBoarderY; }
        if (chLoBoarderX < fTofStationXMin[iStation]) { fTofStationXMin[iStation] = chLoBoarderX; }
        if (chLoBoarderY < fTofStationYMin[iStation]) { fTofStationYMin[iStation] = chLoBoarderY; }

        nTofStationModules[iStation] += 1;
      }
    }
  }

  /// Get the average values
  for (int iSt = 0; iSt < this->GetNtrackingStations(); ++iSt) {
    fTofStationZ[iSt] = fTofStationZ[iSt] / nTofStationModules[iSt];
  }

  // Check the validity of the parameters
  if (!this->Check()) {
    LOG(error)
      << "Some errors occurred in the tracking detector interface initialization for TOF (see information above)";
    return kFATAL;
  }

  return kSUCCESS;
}

//-------------------------------------------------------------------------------------------------------------------------------------
//
InitStatus CbmTofTrackingInterface::ReInit()
{
  this->SetParContainers();
  return Init();
}

//-------------------------------------------------------------------------------------------------------------------------------------
//
void CbmTofTrackingInterface::SetParContainers()
{

  auto runtimeDb = FairRunAna::Instance()->GetRuntimeDb();
  fDigiPar       = dynamic_cast<CbmTofDigiPar*>(runtimeDb->getContainer("CbmTofDigiPar"));
  fDigiBdfPar    = dynamic_cast<CbmTofDigiBdfPar*>(runtimeDb->getContainer("CbmTofDigiBdfPar"));
  if (!fDigiPar) {
    LOG(fatal) << "CbmTofTrackingInterface::SetParContainers: error accessing to CbmTofDigiPar container";
  }
  if (!fDigiBdfPar) {
    LOG(fatal) << "CbmTofTrackingInterface::SetParContainers: error accessing to CbmTofDigiBdfPar container";
  }
  runtimeDb->initContainers(FairRunAna::Instance()->GetRunId());
}
