/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergey Gorbunov [committer] */

#include "CbmKfTrackFitter.h"

#include "CaConstants.h"
#include "CaFramework.h"
#include "CaSimd.h"
#include "CaStation.h"
#include "CaTrackFit.h"
#include "CaTrackParam.h"
#include "CbmGlobalTrack.h"
#include "CbmL1.h"
#include "CbmL1Util.h"
#include "CbmMuchPixelHit.h"
#include "CbmMuchTrack.h"
#include "CbmMuchTrackingInterface.h"
#include "CbmMvdHit.h"
#include "CbmMvdTrackingInterface.h"
#include "CbmStsAddress.h"
#include "CbmStsHit.h"
#include "CbmStsSetup.h"
#include "CbmStsTrack.h"
#include "CbmStsTrackingInterface.h"
#include "CbmTofHit.h"
#include "CbmTofTrack.h"
#include "CbmTofTrackingInterface.h"
#include "CbmTrdHit.h"
#include "CbmTrdTrack.h"
#include "CbmTrdTrackingInterface.h"
#include "FairRootManager.h"
#include "KFParticleDatabase.h"
#include "TClonesArray.h"
#include "TDatabasePDG.h"

using std::vector;
using namespace std;
using ca::fmask;
using ca::fvec;

namespace
{
  using namespace cbm::algo;
}

void CbmKfTrackFitter::Track::MakeConsistent()
{
  // sort the nodes in z
  std::sort(fNodes.begin(), fNodes.end(), [](const FitNode& a, const FitNode& b) { return a.fZ < b.fZ; });

  // set the first and last hit nodes
  fFirstHitNode = fNodes.size() - 1;
  fLastHitNode  = 0;
  for (int i = 0; i < (int) fNodes.size(); i++) {
    if (fNodes[i].fIsXySet) {
      fFirstHitNode = std::min(fFirstHitNode, i);
      fLastHitNode  = std::max(fLastHitNode, i);
    }
  }
}


CbmKfTrackFitter::CbmKfTrackFitter() {}

CbmKfTrackFitter::~CbmKfTrackFitter() {}

void CbmKfTrackFitter::Init()
{
  if (fIsInitialized) {
    return;
  }

  if (!CbmL1::Instance() || !CbmL1::Instance()->fpAlgo) {
    LOG(fatal) << "CbmKfTrackFitter: no CbmL1 task initialized ";
  }

  FairRootManager* ioman = FairRootManager::Instance();

  if (!ioman) {
    LOG(fatal) << "CbmKfTrackFitter: no FairRootManager";
  }

  // Get hits

  fInputMvdHits  = dynamic_cast<TClonesArray*>(ioman->GetObject("MvdHit"));
  fInputStsHits  = dynamic_cast<TClonesArray*>(ioman->GetObject("StsHit"));
  fInputTrdHits  = dynamic_cast<TClonesArray*>(ioman->GetObject("TrdHit"));
  fInputMuchHits = dynamic_cast<TClonesArray*>(ioman->GetObject("MuchHit"));
  fInputTofHits  = dynamic_cast<TClonesArray*>(ioman->GetObject("TofHit"));

  // Get global tracks
  fInputGlobalTracks = dynamic_cast<TClonesArray*>(ioman->GetObject("GlobalTrack"));

  // Get detector tracks
  fInputStsTracks  = dynamic_cast<TClonesArray*>(ioman->GetObject("StsTrack"));
  fInputMuchTracks = dynamic_cast<TClonesArray*>(ioman->GetObject("MuchTrack"));
  fInputTrdTracks  = dynamic_cast<TClonesArray*>(ioman->GetObject("TrdTrack"));
  fInputTofTracks  = dynamic_cast<TClonesArray*>(ioman->GetObject("TofTrack"));

  fIsInitialized = true;
}

void CbmKfTrackFitter::SetParticleHypothesis(int pdg)
{
  TParticlePDG* particlePDG = TDatabasePDG::Instance()->GetParticle(pdg);
  if (!particlePDG) {
    LOG(fatal) << "CbmKfTrackFitter: particle PDG " << pdg << " is not in the data base, please set the mass manually";
    return;
  }
  fMass       = particlePDG->Mass();
  fIsElectron = (abs(pdg) == 11);
}

void CbmKfTrackFitter::SetMassHypothesis(double mass)
{
  assert(mass >= 0.);
  fMass = mass;
}


void CbmKfTrackFitter::SetElectronFlag(bool isElectron) { fIsElectron = isElectron; }


bool CbmKfTrackFitter::CreateMvdStsTrack(CbmKfTrackFitter::Track& kfTrack, int stsTrackIndex)
{
  Init();
  if (!fIsInitialized) {
    return false;
  }

  if (!fInputStsTracks) {
    LOG(error) << "CbmKfTrackFitter: Sts track array not found!";
    return false;
  }
  if (stsTrackIndex >= fInputStsTracks->GetEntriesFast()) {
    LOG(error) << "CbmKfTrackFitter: Sts track index " << stsTrackIndex << " is out of range!";
    return false;
  }
  auto* stsTrack = dynamic_cast<const CbmStsTrack*>(fInputStsTracks->At(stsTrackIndex));
  if (!stsTrack) {
    LOG(error) << "CbmKfTrackFitter: Sts track is null!";
    return false;
  }

  CbmGlobalTrack globalTrack;
  globalTrack.SetStsTrackIndex(stsTrackIndex);
  globalTrack.SetParamFirst(*dynamic_cast<const FairTrackParam*>(stsTrack->GetParamFirst()));

  return CreateGlobalTrack(kfTrack, globalTrack);
}


bool CbmKfTrackFitter::CreateGlobalTrack(CbmKfTrackFitter::Track& kfTrack, int globalTrackIndex)
{
  Init();
  if (!fIsInitialized) {
    return false;
  }

  if (!fInputGlobalTracks) {
    LOG(error) << "CbmKfTrackFitter: Global track array not found!";
    return false;
  }

  if (globalTrackIndex >= fInputGlobalTracks->GetEntriesFast()) {
    LOG(error) << "CbmKfTrackFitter: Global track index " << globalTrackIndex << " is out of range!";
    return false;
  }

  auto* globalTrack = dynamic_cast<const CbmGlobalTrack*>(fInputGlobalTracks->At(globalTrackIndex));
  if (!globalTrack) {
    LOG(error) << "CbmKfTrackFitter: Global track is null!";
    return false;
  }

  return CreateGlobalTrack(kfTrack, *globalTrack);
}


bool CbmKfTrackFitter::CreateGlobalTrack(CbmKfTrackFitter::Track& kfTrack, const CbmGlobalTrack& globalTrack)
{
  kfTrack = {};
  Init();
  if (!fIsInitialized) {
    return false;
  }

  const ca::Parameters<ca::fvec>& caPar = CbmL1::Instance()->fpAlgo->GetParameters();

  int nStations = caPar.GetNstationsActive();

  CbmKfTrackFitter::Track t;

  {
    t.fNodes.resize(nStations);
    for (int i = 0; i < nStations; i++) {
      CbmKfTrackFitter::FitNode& n = t.fNodes[i];
      n                            = {};
      n.fMaterialLayer             = i;
      n.fZ                         = caPar.GetStation(i).GetZ<float>();
      n.fRadThick                  = 0.;
      n.fIsRadThickSet             = false;
      n.fIsFitted                  = false;
      n.fIsXySet                   = false;
      n.fIsTimeSet                 = false;
      n.fHitSystemId               = ECbmModuleId::kNotExist;
      n.fHitAddress                = -1;
      n.fHitIndex                  = -1;
    }
    t.fFirstHitNode = nStations - 1;
    t.fLastHitNode  = 0;
  }

  // lambda to set the node from the pixel hit
  auto setNode = [&](const CbmPixelHit& h, int stIdx, int hitIdx, ca::EDetectorID detId) {
    int ista = caPar.GetStationIndexActive(stIdx, detId);
    if (ista < 0) {
      return;
    }
    assert(ista < nStations);
    CbmKfTrackFitter::FitNode& n = t.fNodes[ista];
    n.fZ                         = h.GetZ();
    n.fMxy.SetX(h.GetX());
    n.fMxy.SetY(h.GetY());
    n.fMxy.SetDx2(h.GetDx() * h.GetDx());
    n.fMxy.SetDy2(h.GetDy() * h.GetDy());
    n.fMxy.SetDxy(h.GetDxy());
    n.fMxy.SetNdfX(1);
    n.fMxy.SetNdfY(1);
    n.fIsXySet = true;
    n.fMt.SetT(h.GetTime());
    n.fMt.SetDt2(h.GetTimeError() * h.GetTimeError());
    n.fMt.SetNdfT(1);
    n.fIsTimeSet     = (detId != ca::EDetectorID::kMvd);
    n.fRadThick      = 0.;
    n.fIsRadThickSet = false;
    n.fHitSystemId   = cbm::L1Util::ConvertDetSystemId(detId);
    n.fHitAddress    = h.GetAddress();
    n.fHitIndex      = hitIdx;
  };

  // Read MVD & STS hits

  if (globalTrack.GetStsTrackIndex() >= 0) {

    int stsTrackIndex = globalTrack.GetStsTrackIndex();

    if (!fInputStsTracks) {
      LOG(error) << "CbmKfTrackFitter: Sts track array not found!";
      return false;
    }
    if (stsTrackIndex >= fInputStsTracks->GetEntriesFast()) {
      LOG(error) << "CbmKfTrackFitter: Sts track index " << stsTrackIndex << " is out of range!";
      return false;
    }
    auto* stsTrack = dynamic_cast<const CbmStsTrack*>(fInputStsTracks->At(stsTrackIndex));
    if (!stsTrack) {
      LOG(error) << "CbmKfTrackFitter: Sts track is null!";
      return false;
    }

    // Read MVD hits

    int nMvdHits = stsTrack->GetNofMvdHits();
    if (nMvdHits > 0) {
      if (!fInputMvdHits) {
        LOG(error) << "CbmKfTrackFitter: Mvd hit array not found!";
        return false;
      }
      for (int ih = 0; ih < nMvdHits; ih++) {
        int hitIndex = stsTrack->GetMvdHitIndex(ih);
        if (hitIndex >= fInputMvdHits->GetEntriesFast()) {
          LOG(error) << "CbmKfTrackFitter: Mvd hit index " << hitIndex << " is out of range!";
          return false;
        }
        const auto& hit = *dynamic_cast<const CbmMvdHit*>(fInputMvdHits->At(hitIndex));
        setNode(hit, CbmMvdTrackingInterface::Instance()->GetTrackingStationIndex(&hit), hitIndex,
                ca::EDetectorID::kMvd);
      }
    }

    // Read STS hits

    int nStsHits = stsTrack->GetNofStsHits();
    if (nStsHits > 0) {
      if (!fInputStsHits) {
        LOG(error) << "CbmKfTrackFitter: Sts hit array not found!";
        return false;
      }
      for (int ih = 0; ih < nStsHits; ih++) {
        int hitIndex = stsTrack->GetStsHitIndex(ih);
        if (hitIndex >= fInputStsHits->GetEntriesFast()) {
          LOG(error) << "CbmKfTrackFitter: Sts hit index " << hitIndex << " is out of range!";
          return false;
        }
        const auto& hit = *dynamic_cast<const CbmStsHit*>(fInputStsHits->At(hitIndex));
        setNode(hit, CbmStsTrackingInterface::Instance()->GetTrackingStationIndex(&hit), hitIndex,
                ca::EDetectorID::kSts);
      }
    }
  }  // MVD & STS hits


  // Read Much hits

  if (globalTrack.GetMuchTrackIndex() >= 0) {
    int muchTrackIndex = globalTrack.GetMuchTrackIndex();
    if (!fInputMuchTracks) {
      LOG(error) << "CbmKfTrackFitter: Much track array not found!";
      return false;
    }
    if (muchTrackIndex >= fInputMuchTracks->GetEntriesFast()) {
      LOG(error) << "CbmKfTrackFitter: Much track index " << muchTrackIndex << " is out of range!";
      return false;
    }
    auto* track = dynamic_cast<const CbmMuchTrack*>(fInputMuchTracks->At(muchTrackIndex));
    if (!track) {
      LOG(error) << "CbmKfTrackFitter: Much track is null!";
      return false;
    }
    int nHits = track->GetNofHits();
    if (nHits > 0) {
      if (!fInputMuchHits) {
        LOG(error) << "CbmKfTrackFitter: Much hit array not found!";
        return false;
      }
      for (int ih = 0; ih < nHits; ih++) {
        int hitIndex = track->GetHitIndex(ih);
        if (hitIndex >= fInputMuchHits->GetEntriesFast()) {
          LOG(error) << "CbmKfTrackFitter: Much hit index " << hitIndex << " is out of range!";
          return false;
        }
        const auto& hit = *dynamic_cast<const CbmMuchPixelHit*>(fInputMuchHits->At(hitIndex));
        setNode(hit, CbmMuchTrackingInterface::Instance()->GetTrackingStationIndex(&hit), hitIndex,
                ca::EDetectorID::kMuch);
      }
    }
  }

  // Read TRD hits

  if (globalTrack.GetTrdTrackIndex() >= 0) {
    int trdTrackIndex = globalTrack.GetTrdTrackIndex();
    if (!fInputTrdTracks) {
      LOG(error) << "CbmKfTrackFitter: Trd track array not found!";
      return false;
    }
    if (trdTrackIndex >= fInputTrdTracks->GetEntriesFast()) {
      LOG(error) << "CbmKfTrackFitter: Trd track index " << trdTrackIndex << " is out of range!";
      return false;
    }
    auto* track = dynamic_cast<const CbmTrdTrack*>(fInputTrdTracks->At(trdTrackIndex));
    if (!track) {
      LOG(error) << "CbmKfTrackFitter: Trd track is null!";
      return false;
    }
    int nHits = track->GetNofHits();
    if (nHits > 0) {
      if (!fInputTrdHits) {
        LOG(error) << "CbmKfTrackFitter: Trd hit array not found!";
        return false;
      }
      for (int ih = 0; ih < nHits; ih++) {
        int hitIndex = track->GetHitIndex(ih);
        if (hitIndex >= fInputTrdHits->GetEntriesFast()) {
          LOG(error) << "CbmKfTrackFitter: Trd hit index " << hitIndex << " is out of range!";
          return false;
        }
        const auto& hit = *dynamic_cast<const CbmTrdHit*>(fInputTrdHits->At(hitIndex));
        setNode(hit, CbmTrdTrackingInterface::Instance()->GetTrackingStationIndex(&hit), hitIndex,
                ca::EDetectorID::kTrd);
      }
    }
  }


  // Read TOF hits

  if (globalTrack.GetTofTrackIndex() >= 0) {
    int tofTrackIndex = globalTrack.GetTofTrackIndex();
    if (!fInputTofTracks) {
      LOG(error) << "CbmKfTrackFitter: Trd track array not found!";
      return false;
    }
    if (tofTrackIndex >= fInputTofTracks->GetEntriesFast()) {
      LOG(error) << "CbmKfTrackFitter: Trd track index " << tofTrackIndex << " is out of range!";
      return false;
    }
    auto* track = dynamic_cast<const CbmTofTrack*>(fInputTofTracks->At(tofTrackIndex));
    if (!track) {
      LOG(error) << "CbmKfTrackFitter: Tof track is null!";
      return false;
    }

    int nHits = track->GetNofHits();
    if (nHits > 0) {
      if (!fInputTofHits) {
        LOG(error) << "CbmKfTrackFitter: Tof hit array not found!";
        return false;
      }
      for (int ih = 0; ih < nHits; ih++) {
        int hitIndex = track->GetHitIndex(ih);
        if (hitIndex >= fInputTofHits->GetEntriesFast()) {
          LOG(error) << "CbmKfTrackFitter: Tof hit index " << hitIndex << " is out of range!";
          return false;
        }
        const auto& hit = *dynamic_cast<const CbmTofHit*>(fInputTofHits->At(hitIndex));
        setNode(hit, CbmTofTrackingInterface::Instance()->GetTrackingStationIndex(&hit), hitIndex,
                ca::EDetectorID::kTof);
      }
    }
  }

  t.MakeConsistent();

  ca::TrackParamD tmp = cbm::L1Util::ConvertTrackParam(*globalTrack.GetParamFirst());

  t.fNodes[t.fFirstHitNode].fTrack.Set(tmp);
  t.fNodes[t.fFirstHitNode].fIsFitted = 1;

  kfTrack = t;
  return true;

  //  return CreateTrack(kfTrack, *globalTrack.GetParamFirst(), mvdHits, stsHits, muchHits, trdHits, tofHits);
}


void CbmKfTrackFitter::FilterFirstMeasurement(const FitNode& n)
{
  // a special routine to filter the first measurement.
  // the measurement errors are simply copied to the track covariance matrix

  const auto& mxy = n.fMxy;
  const auto& mt  = n.fMt;

  auto& tr = fFit.Tr();
  tr.ResetErrors(mxy.Dx2(), mxy.Dy2(), 1., 1., 1., 1.e4, 1.e2);
  tr.SetC10(mxy.Dxy());
  tr.SetX(mxy.X());
  tr.SetY(mxy.Y());
  tr.SetChiSq(0.);
  tr.SetChiSqTime(0.);
  tr.SetNdf(-5 + mxy.NdfX() + mxy.NdfY());
  if (n.fIsTimeSet) {
    tr.SetTime(mt.T());
    tr.SetC55(mt.Dt2());
    tr.SetNdfTime(-2 + 1);
  }
  else {
    tr.SetNdfTime(-2);
  }
  tr.SetVi(0.);
}


void CbmKfTrackFitter::AddMaterialEffects(const CbmKfTrackFitter::Track& t, CbmKfTrackFitter::FitNode& n,
                                          bool upstreamDirection)
{
  // add material effects
  if (n.fMaterialLayer < 0) {
    return;
  }

  auto& tr = n.fIsFitted ? n.fTrack : fFit.Tr();

  // calculate the radiation thickness from the current track
  if (!n.fIsRadThickSet) {
    n.fRadThick =
      CbmL1::Instance()->fpAlgo->GetParameters().GetMaterialThicknessScal(n.fMaterialLayer, tr.GetX()[0], tr.GetY()[0]);
  }

  fvec msQp0 = t.fIsMsQp0Set ? t.fMsQp0 : tr.GetQp();

  fFit.MultipleScattering(n.fRadThick, msQp0);
  fFit.EnergyLossCorrection(n.fRadThick, upstreamDirection ? fvec::One() : fvec::Zero());
}


void CbmKfTrackFitter::FitTrack(CbmKfTrackFitter::Track& t)
{
  // fit the track

  // ensure that the fitter is initialized
  Init();

  t.MakeConsistent();

  fFit.SetMask(fmask::One());
  fFit.SetParticleMass(fMass);

  ca::FieldRegion<ca::fvec> field _fvecalignment;
  field.SetUseOriginalField();

  int nNodes = t.fNodes.size();

  // fit downstream. The approximation is taken from the first hit node
  {
    FitNode& n = t.fNodes[t.fFirstHitNode];
    fFit.SetTrack(n.fTrack);
    FilterFirstMeasurement(n);
    n.fTrack    = fFit.Tr();
    n.fIsFitted = false;
  }

  for (int iNode = t.fFirstHitNode + 1; iNode < nNodes; iNode++) {
    FitNode& n = t.fNodes[iNode];
    fFit.Extrapolate(n.fZ, field);
    if (n.fIsFitted) {
      fFit.SetQp0(n.fTrack.GetQp());
    }
    AddMaterialEffects(t, n, false);

    // store track fitted at the previous node and extrapolated to this node
    n.fTrack    = fFit.Tr();
    n.fIsFitted = false;
    if (n.fIsXySet) {
      fFit.FilterXY(n.fMxy);
    }
    if (n.fIsTimeSet) {
      fFit.FilterTime(n.fMt);
    }
    if (iNode >= t.fLastHitNode) {  // store track fitted at the last node
      n.fTrack    = fFit.Tr();
      n.fIsFitted = true;
    }
  }

  // fit upstream
  {
    FitNode& n = t.fNodes[t.fLastHitNode];
    fFit.SetTrack(n.fTrack);
    FilterFirstMeasurement(n);
    n.fIsFitted = true;
  }

  for (int iNode = t.fLastHitNode - 1; iNode >= 0; iNode--) {
    FitNode& n = t.fNodes[iNode];
    fFit.Extrapolate(n.fZ, field);

    if (n.fIsXySet) {
      fFit.FilterXY(n.fMxy);
    }
    if (n.fIsTimeSet) {
      fFit.FilterTime(n.fMt);
    }

    // combine partially fitted downstream and upstream tracks
    if (iNode > t.fFirstHitNode) {
      Smooth(n.fTrack, fFit.Tr());
    }
    else {
      n.fTrack = fFit.Tr();
    }
    n.fIsFitted = true;
    fFit.SetQp0(n.fTrack.GetQp());
    AddMaterialEffects(t, n, true);
  }

  // distribute the final chi2, ndf to all nodes

  const auto& tt = t.fNodes[t.fFirstHitNode].fTrack;
  for (auto& n : t.fNodes) {
    n.fTrack.SetNdf(tt.GetNdf());
    n.fTrack.SetNdfTime(tt.GetNdfTime());
    n.fTrack.SetChiSq(tt.GetChiSq());
    n.fTrack.SetChiSqTime(tt.GetChiSqTime());
  }
}


void CbmKfTrackFitter::Smooth(ca::TrackParamV& t1, const ca::TrackParamV& t2)
{
  // combine two tracks
  std::tie(t1.X(), t1.Y(), t1.C00(), t1.C10(), t1.C11()) =
    Smooth2D(t1.X(), t1.Y(), t1.C00(), t1.C10(), t1.C11(), t2.X(), t2.Y(), t2.C00(), t2.C10(), t2.C11());

  std::tie(t1.Tx(), t1.Ty(), t1.C22(), t1.C32(), t1.C33()) =
    Smooth2D(t1.Tx(), t1.Ty(), t1.C22(), t1.C32(), t1.C33(), t2.Tx(), t2.Ty(), t2.C22(), t2.C32(), t2.C33());

  std::tie(t1.Qp(), t1.C44()) = Smooth1D(t1.Qp(), t1.C44(), t2.Qp(), t2.C44());

  std::tie(t1.Time(), t1.Vi(), t1.C55(), t1.C65(), t1.C66()) =
    Smooth2D(t1.Time(), t1.Vi(), t1.C55(), t1.C65(), t1.C66(), t2.Time(), t2.Vi(), t2.C55(), t2.C65(), t2.C66());

  t1.C20() = 0.;
  t1.C21() = 0.;
  t1.C30() = 0.;
  t1.C31() = 0.;
  t1.C40() = 0.;
  t1.C41() = 0.;
  t1.C42() = 0.;
  t1.C43() = 0.;
  t1.C50() = 0.;
  t1.C51() = 0.;
  t1.C52() = 0.;
  t1.C53() = 0.;
  t1.C54() = 0.;
  t1.C60() = 0.;
  t1.C61() = 0.;
  t1.C62() = 0.;
  t1.C63() = 0.;
  t1.C64() = 0.;
}

std::tuple<ca::fvec, ca::fvec> CbmKfTrackFitter::Smooth1D(ca::fvec x1, ca::fvec Cxx1, ca::fvec x2, ca::fvec Cxx2)
{
  // combine two 1D values
  ca::fvec x   = (x1 * Cxx1 + x2 * Cxx2) / (Cxx1 + Cxx2);
  ca::fvec Cxx = Cxx1 * Cxx2 / (Cxx1 + Cxx2);
  return std::tuple(x, Cxx);
}

std::tuple<ca::fvec, ca::fvec, ca::fvec, ca::fvec, ca::fvec>
CbmKfTrackFitter::Smooth2D(ca::fvec x1, ca::fvec y1, ca::fvec Cxx1, ca::fvec /*Cxy1*/, ca::fvec Cyy1, ca::fvec x2,
                           ca::fvec y2, ca::fvec Cxx2, ca::fvec /*Cxy2*/, ca::fvec Cyy2)
{
  // combine two 2D values
  // TODO: do it right
  auto [x, Cxx] = Smooth1D(x1, Cxx1, x2, Cxx2);
  auto [y, Cyy] = Smooth1D(y1, Cyy1, y2, Cyy2);
  return std::tuple(x, y, Cxx, ca::fvec::Zero(), Cyy);
}
