/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// @file   CbmCaInputQaMvd.cxx
/// @date   10.09.2023
/// @brief  QA-task for CA tracking input from MVD (implementation)
/// @author S.Zharko <s.zharko@gsi.de>

#include "CbmCaInputQaMvd.h"

#include "CbmMvdTrackingInterface.h"

ClassImp(CbmCaInputQaMvd);

// ---------------------------------------------------------------------------------------------------------------------
//
CbmCaInputQaMvd::CbmCaInputQaMvd(int verbose, bool isMCUsed) : CbmCaInputQaBase("CbmCaInputQaMvd", verbose, isMCUsed)
{
  // Default parameters of task
  DefineParameters();
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmCaInputQaMvd::DeInit() { CbmCaInputQaBase::DeInit(); }

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmCaInputQaMvd::DefineParameters()
{
  auto SetRange = [](std::array<double, 2>& range, double min, double max) {
    range[0] = min;
    range[1] = max;
  };
  // Hit errors
  SetRange(fRHitDx, -0.005, 0.005);  // [cm]
  SetRange(fRHitDy, -0.005, 0.005);  // [cm]
  SetRange(fRHitDu, -0.005, 0.005);  // [cm]
  SetRange(fRHitDv, -0.005, 0.005);  // [cm]
  SetRange(fRHitDt, -10., 10.);      // [ns]
  // Residuals
  SetRange(fRResX, -0.004, 0.004);
  SetRange(fRResY, -0.004, 0.004);
  SetRange(fRResU, -0.004, 0.004);
  SetRange(fRResV, -0.004, 0.004);
  SetRange(fRResT, -5.0, 5.0);
  // QA result selection criteria
  SetRange(fEffRange, 10.0, 30.0);  ///< Range for hit efficiency approximation
  fResMeanThrsh   = 0.50;           ///< Maximum allowed deviation of residual mean from zero [sigma]
  fPullMeanThrsh  = 0.10;           ///< Maximum allowed deviation of pull mean from zero
  fPullWidthThrsh = 2.00;           ///< Maximum allowed deviation of pull width from unity
  fEffThrsh       = 0.50;           ///< Threshold for hit efficiency in the selected range
  fMaxDiffZStHit  = 1.00;           ///< Maximum allowed difference between z-position of hit and station [cm]
  fMinMomentum    = 0.05;           ///< Minimum momentum of particle [GeV/c]
  // Track selection criteria
  fTrackSelectionPrimary  = true;  ///< must the track come from the primary vertex
  fTrackSelectionMomentum = 0.1;   ///< track momentum GeV when it exits the tracking station
  fTrackSelectionAngle    = 60.;   ///< track incl. angle [grad] when it exits the station
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmCaInputQaMvd::FillHistogramsPerHit() {}

// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus CbmCaInputQaMvd::InitDataBranches()
{
  // STS tracking detector interface
  fpDetInterface = CbmMvdTrackingInterface::Instance();

  auto baseInitStatus = CbmCaInputQaBase::InitDataBranches();
  if (kSUCCESS != baseInitStatus) {
    return baseInitStatus;
  }

  return kSUCCESS;
}

// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus CbmCaInputQaMvd::InitCanvases()
{
  auto baseInitStatus = CbmCaInputQaBase::InitCanvases();
  if (kSUCCESS != baseInitStatus) {
    return baseInitStatus;
  }

  return kSUCCESS;
}

// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus CbmCaInputQaMvd::InitHistograms()
{
  auto baseInitStatus = CbmCaInputQaBase::InitHistograms();
  if (kSUCCESS != baseInitStatus) {
    return baseInitStatus;
  }

  return kSUCCESS;
}
