/* Copyright (C) 2023 Facility for Antiproton and Ion Research in Europe, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Pierre-Alain Loizeau [committer] */

#include "Application.h"

#include "CbmFlesCanvasTools.h"

#include <Logger.h>

#include <boost/archive/binary_oarchive.hpp>
#include <boost/iostreams/device/array.hpp>
#include <boost/iostreams/device/back_inserter.hpp>
#include <boost/iostreams/stream.hpp>
#include <boost/serialization/utility.hpp>
#include <boost/serialization/vector.hpp>

#include <thread>

#include "Histo1D.h"
#include "ui_callbacks.h"

std::mutex mtx;

namespace b_io = boost::iostreams;
namespace b_ar = boost::archive;
using sctp     = std::chrono::time_point<std::chrono::system_clock>;
using scsc     = std::chrono::system_clock;

namespace cbm::services::histserv_tester
{

  // -----   Constructor   ---------------------------------------------------------------------------------------------
  Application::Application(ProgramOptions const& opt) : fOpt(opt)
  {
    /// Read options from executable
    LOG(info) << "Options for Application.";
    LOG(info) << " Output ZMQ channel: " << fOpt.ComChan();
    LOG(info) << " Run time duration:  " << fOpt.Runtime() << " s";

    /// FIXME: SOMETHING_To_Replace_FairMQ!!!!!!!!!!!!!
    /// FIXME: Initialize communication channels of SOMETHING_To_Replace_FairMQ
    /// FIXME: Link channel to method in order to process received messages
    // fZmqSocket.set(zmq::sockopt::rcvhwm, int(hwm));  // FIXME: need for HWM?
    fZmqSocket.connect(fOpt.ComChan().c_str());  // This side "connects" to socket => Other side should have "bind"!!!!
  }
  // -------------------------------------------------------------------------------------------------------------------


  // -----   Main Loop   -----------------------------------------------------------------------------------------------
  void Application::Exec()
  {
    const std::chrono::milliseconds interval {250};
    const std::chrono::seconds runtime(fOpt.Runtime());
    const std::chrono::seconds pubint(fOpt.PubInterval());
    sctp startTime   = scsc::now();
    sctp stopTime    = startTime + runtime;
    sctp lastPubTime = startTime;

    std::vector<cbm::algo::Histo1D> vHist;
    vHist.emplace_back((fOpt.Runtime() + 2), -1.0, fOpt.Runtime() + 1.0, "testHist",
                       "Tester source; Runtime [s]; Entries to histo [iterations]");
    vHist.emplace_back(1001, -0.025, 50.025, "transHist",
                       "Tester histos transmission time; Trans. time [ms]; Messages []");

    /// Initial emission, including generation and serialization of configs
    /// => Try to evaluate "time cost" of the histo transmission, including serialization
    sctp transStartTime = scsc::now();

    /// => Header for multi-part message with Configuration + data
    /// => Format: std::pair< Nb histogram configs, Nb canvas configs  >
    std::pair<uint32_t, uint32_t> pHeader(vHist.size(), 1);
    PrepareAndSendMsg(pHeader, zmq::send_flags::sndmore);

    /// => Histograms configuration = destination folder in http browser, mandatory but can be empty (= root folder)
    /// => 1 ZMQ message per histogram (= 1 part)
    /// => If no (new) histograms declared (e.g. new canvas declaration), has to be en empty message + `0` in the header
    std::pair<std::string, std::string> pFirstHist(vHist[0].Name(), "TestFolder");
    std::pair<std::string, std::string> pSecondHist(vHist[1].Name(), "TestFolder/TestSubFolder");
    PrepareAndSendMsg(pFirstHist, zmq::send_flags::sndmore);
    PrepareAndSendMsg(pSecondHist, zmq::send_flags::sndmore);

    /// => Canvas configuration = in old code extracted from canvas object, here need to be made by hand
    /// => 1 ZMQ message per canvas (= 1 part)
    /// => If no (new) canvas declared (e.g. only histos declaration), has to be en empty message + `0` in the header
    /// => Format is "CanvasName;Canvas Title;NbPadX(U);NbPadY(U);ConfigPad1(s);....;ConfigPadXY(s)"
    /// => Format of Pad config is
    ///    "GrixX(b),GridY(b),LogX(b),LogY(b),LogZ(b),(HistoName1,DrawOptions1),...,(HistoNameZ,DrawOptionsZ)"
    /// => See core/base/utils/fles/CbmFlesCanvasTools for the full code, especially GenerateCanvasConfigString
    ///    (CanvasConfig class not used here as library loading would need binding to ROOT library)
    std::pair<std::string, std::string> pCanvConfig("TestCanvas", "TestCanvas;Canvas configs testing;2;2;");
    pCanvConfig.second += "0,0,0,0,0,(" + vHist[0].Name() + ",hist);";                                      // Pad 0
    pCanvConfig.second += "1,1,0,0,0,(" + vHist[0].Name() + ",hist);";                                      // Pad 1
    pCanvConfig.second += "1,1,1,0,0,(" + vHist[1].Name() + ",hist);";                                      // Pad 2
    pCanvConfig.second += "1,1,1,1,0,(" + vHist[1].Name() + ",hist),(" + vHist[0].Name() + ",hist same);";  // Pad 3
    PrepareAndSendMsg(pCanvConfig, zmq::send_flags::sndmore);

    /// => (empty) Histograms serialization and emission
    PrepareAndSendMsg(vHist, zmq::send_flags::none);
    lastPubTime = scsc::now();

    /// No general references as member/variable bec. simple example, use directly hardcoded vector "array access"
    vHist[1].Add(std::chrono::duration_cast<std::chrono::microseconds>(lastPubTime - transStartTime).count() / 1e3);

    while (scsc::now() < stopTime) {  //
      /// No general references as member/variable bec. simple example, use directly hardcoded vector "array access"
      vHist[0].Add(std::chrono::duration_cast<std::chrono::milliseconds>(scsc::now() - startTime).count() / 1e3);

      if (pubint < scsc::now() - lastPubTime) {
        LOG(info) << "Pub-hist: " << scsc::now().time_since_epoch().count() << " " << vHist[0].NumEntries();
        /// Try to evaluate "time cost" of the histo transmission, including serialization
        transStartTime = scsc::now();

        /// => Histograms serialization and emission
        PrepareAndSendMsg(vHist, zmq::send_flags::none);

        /// => Reset all histograms as they are added on the other end!
        for (std::vector<cbm::algo::Histo1D>::iterator itHist = vHist.begin(); itHist != vHist.end(); ++itHist) {

          itHist->Clear();
        }
        lastPubTime = scsc::now();

        /// No general references as member/variable bec. simple example, use directly hardcoded vector "array access"
        vHist[1].Add(std::chrono::duration_cast<std::chrono::microseconds>(lastPubTime - transStartTime).count() / 1e3);
      }

      std::this_thread::sleep_for(interval);
      LOG(info) << "test: " << (scsc::now() < stopTime) << " (" << scsc::now().time_since_epoch().count() << " vs "
                << stopTime.time_since_epoch().count() << ") => "
                << (std::chrono::duration_cast<std::chrono::milliseconds>(scsc::now() - startTime).count() / 1e3);
    }

    /// Final publications
    LOG(info) << "Pub-hist: " << scsc::now().time_since_epoch().count() << " " << vHist[0].NumEntries();
    /// => Histograms serialization and emission
    PrepareAndSendMsg(vHist, zmq::send_flags::none);
    /// => Reset all histograms as they are added on the other end!
    for (std::vector<cbm::algo::Histo1D>::iterator itHist = vHist.begin(); itHist != vHist.end(); ++itHist) {
      itHist->Clear();
    }
  }
  // -------------------------------------------------------------------------------------------------------------------

  template<class Object>
  void Application::PrepareAndSendMsg(Object& obj, zmq::send_flags flags)
  {
    /// Needed ressources (serializd string, boost inserter, boost stream, boost binary output archive)
    std::string serial_str;
    b_io::back_insert_device<std::string> inserter(serial_str);
    b_io::stream<b_io::back_insert_device<std::string>> bstream(inserter);
    b_ar::binary_oarchive oa(bstream);

    serial_str.clear();
    oa << obj;
    bstream.flush();

    zmq::message_t msg(serial_str.size());
    std::copy_n(static_cast<const char*>(serial_str.data()), msg.size(), static_cast<char*>(msg.data()));
    fZmqSocket.send(msg, flags);
  }

}  // namespace cbm::services::histserv_tester
